# -*- coding: utf-8 -*-
"""Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1JdXHhEyjjEE--19ZPH1bZV_LiGQBndzs

Copyright (c) 2021 Björn Ottosson

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

**Overview**

This notebook was used to compute coefficients for the compute_max_saturation function in this blog post:
http://bottosson.github.io/posts/gamutclipping/

The code is available for reference, since it could be useful to derive coefficients for other color spaces. It was
written quickly to derive the values and both structure and documentation is poor.
"""

# Commented out `IPython` magic to ensure Python compatibility.
import numpy as np
import scipy.optimize
import matplotlib.pyplot as plt
import sys
import os

sys.path.insert(0, os.getcwd())

# Use higher precision Oklab conversion matrix along with LMS matrix with our exact white point
from tools.calc_oklab_matrices import SRGBL_TO_LMS, LMS_TO_SRGBL, LMS3_TO_OKLAB, OKLAB_TO_LMS3, XYZ_TO_LMS # noqa: E402

PRINT_DIAGS = False
GAMUT = 'srgb'
np.set_printoptions(precision=8)

# Use P3 gamut (or some other gamut)
# Calculate the gamut <-> LMS matrices to adjust the working gamut
if GAMUT == 'p3':
    import coloraide.spaces.display_p3_linear as p3
    from coloraide import algebra as alg

    RGBL_TO_LMS = alg.dot(XYZ_TO_LMS, p3.RGB_TO_XYZ)
    LMS_TO_RGBL = alg.inv(RGBL_TO_LMS)
else:
    RGBL_TO_LMS = SRGBL_TO_LMS
    LMS_TO_RGBL = LMS_TO_SRGBL

print(RGBL_TO_LMS)
print(LMS_TO_RGBL)

RGBL_TO_LMS = np.asfarray(RGBL_TO_LMS)
LMS_TO_RGBL = np.asfarray(LMS_TO_RGBL)
LMS3_TO_OKLAB = np.asfarray(LMS3_TO_OKLAB)
OKLAB_TO_LMS3 = np.asfarray(OKLAB_TO_LMS3)


def linear_srgb_to_oklab(c):
    l = RGBL_TO_LMS[0][0] * c[0,...] + RGBL_TO_LMS[0][1] * c[1,...] + RGBL_TO_LMS[0][2] * c[2,...]
    m = RGBL_TO_LMS[1][0] * c[0,...] + RGBL_TO_LMS[1][1] * c[1,...] + RGBL_TO_LMS[1][2] * c[2,...]
    s = RGBL_TO_LMS[2][0] * c[0,...] + RGBL_TO_LMS[2][1] * c[1,...] + RGBL_TO_LMS[2][2] * c[2,...]

    l_ = np.cbrt(l)
    m_ = np.cbrt(m)
    s_ = np.cbrt(s)

    return np.array([
        LMS3_TO_OKLAB[0][0] * l_ + LMS3_TO_OKLAB[0][1] * m_ + LMS3_TO_OKLAB[0][2] * s_,
        LMS3_TO_OKLAB[1][0] * l_ + LMS3_TO_OKLAB[1][1] * m_ + LMS3_TO_OKLAB[1][2] * s_,
        LMS3_TO_OKLAB[2][0] * l_ + LMS3_TO_OKLAB[2][1] * m_ + LMS3_TO_OKLAB[2][2] * s_,
    ])

# define functions for R, G and B as functions of S,h (with L = 1 and S = C/L)

def to_lms(S,h):
  a = S*np.cos(h)
  b = S*np.sin(h)

  l_ = OKLAB_TO_LMS3[0][0] + OKLAB_TO_LMS3[0][1] * a + OKLAB_TO_LMS3[0][2] * b
  m_ = OKLAB_TO_LMS3[1][0] + OKLAB_TO_LMS3[1][1] * a + OKLAB_TO_LMS3[1][2] * b
  s_ = OKLAB_TO_LMS3[2][0] + OKLAB_TO_LMS3[2][1] * a + OKLAB_TO_LMS3[2][2] * b

  l = l_*l_*l_
  m = m_*m_*m_
  s = s_*s_*s_

  return (l,m,s)

def to_lms_dS(S,h):
  a = S*np.cos(h)
  b = S*np.sin(h)

  l_ = OKLAB_TO_LMS3[0][0] + OKLAB_TO_LMS3[0][1] * a + OKLAB_TO_LMS3[0][2] * b
  m_ = OKLAB_TO_LMS3[1][0] + OKLAB_TO_LMS3[1][1] * a + OKLAB_TO_LMS3[1][2] * b
  s_ = OKLAB_TO_LMS3[2][0] + OKLAB_TO_LMS3[2][1] * a + OKLAB_TO_LMS3[2][2] * b

  l = (LMS3_TO_OKLAB[0][1]*np.cos(h) + LMS3_TO_OKLAB[0][1]*np.sin(h))*3*l_*l_
  m = (LMS3_TO_OKLAB[1][1]*np.cos(h) + LMS3_TO_OKLAB[1][1]*np.sin(h))*3*m_*m_
  s = (LMS3_TO_OKLAB[2][1]*np.cos(h) + LMS3_TO_OKLAB[2][1]*np.sin(h))*3*s_*s_

  return (l,m,s)

def to_lms_dS2(S,h):
  a = S*np.cos(h)
  b = S*np.sin(h)

  l_ = OKLAB_TO_LMS3[0][0] + OKLAB_TO_LMS3[0][1] * a + OKLAB_TO_LMS3[0][2] * b
  m_ = OKLAB_TO_LMS3[1][0] + OKLAB_TO_LMS3[1][1] * a + OKLAB_TO_LMS3[1][2] * b
  s_ = OKLAB_TO_LMS3[2][0] + OKLAB_TO_LMS3[2][1] * a + OKLAB_TO_LMS3[2][2] * b

  l = (LMS3_TO_OKLAB[0][1]*np.cos(h) + LMS3_TO_OKLAB[0][2]*np.sin(h))**2*6*l_
  m = (LMS3_TO_OKLAB[1][1]*np.cos(h) + LMS3_TO_OKLAB[0][2]*np.sin(h))**2*6*m_
  s = (LMS3_TO_OKLAB[2][1]*np.cos(h) + LMS3_TO_OKLAB[0][2]*np.sin(h))**2*6*s_

  return (l,m,s)


def to_R(S,h):
  (l,m,s) = to_lms(S,h)
  return LMS_TO_RGBL[0][0]*l + LMS_TO_RGBL[0][1]*m + LMS_TO_RGBL[0][2]*s

def to_R_dS(S,h):
  (l,m,s) = to_lms_dS(S,h)
  return LMS_TO_RGBL[0][0]*l + LMS_TO_RGBL[0][1]*m + LMS_TO_RGBL[0][2]*s

def to_R_dS2(S,h):
  (l,m,s) = to_lms_dS2(S,h)
  return LMS_TO_RGBL[0][0]*l + LMS_TO_RGBL[0][1]*m + LMS_TO_RGBL[0][2]*s

def to_G(S,h):
  (l,m,s) = to_lms(S,h)
  return LMS_TO_RGBL[1][0]*l + LMS_TO_RGBL[1][1]*m + LMS_TO_RGBL[1][2]*s

def to_G_dS(S,h):
  (l,m,s) = to_lms_dS(S,h)
  return LMS_TO_RGBL[1][0]*l + LMS_TO_RGBL[1][1]*m + LMS_TO_RGBL[1][2]*s

def to_G_dS2(S,h):
  (l,m,s) = to_lms_dS2(S,h)
  return LMS_TO_RGBL[1][0]*l + LMS_TO_RGBL[1][1]*m + LMS_TO_RGBL[1][2]*s

def to_B(S,h):
  (l,m,s) = to_lms(S,h)
  return LMS_TO_RGBL[2][0]*l + LMS_TO_RGBL[2][1]*m + LMS_TO_RGBL[2][2]*s

def to_B_dS(S,h):
  (l,m,s) = to_lms_dS(S,h)
  return LMS_TO_RGBL[2][0]*l + LMS_TO_RGBL[2][1]*m + LMS_TO_RGBL[2][2]*s

def to_B_dS2(S,h):
  (l,m,s) = to_lms_dS2(S,h)
  return LMS_TO_RGBL[2][0]*l + LMS_TO_RGBL[2][1]*m + LMS_TO_RGBL[2][2]*s

hs,Ss = np.meshgrid(np.linspace(-np.pi,np.pi,720),np.linspace(0,1,200))

Rs = to_R(Ss, hs)
Gs = to_G(Ss, hs)
Bs = to_B(Ss, hs)

gamut = np.minimum(Rs, np.minimum(Gs, Bs))

if PRINT_DIAGS:
    plt.imshow(np.sign(gamut), cmap='gray', vmin=0, vmax=1)
    plt.show()
    plt.figure()

r_lab = linear_srgb_to_oklab(np.array([1,0,0]))
g_lab = linear_srgb_to_oklab(np.array([0,1,0]))
b_lab = linear_srgb_to_oklab(np.array([0,0,1]))

r_h = np.arctan2(r_lab[2], r_lab[1])
g_h = np.arctan2(g_lab[2], g_lab[1])
b_h = np.arctan2(b_lab[2], b_lab[1])

print(r_h)
print(g_h)
print(b_h)

r_dir = 0.5*np.array([np.cos(b_h)+np.cos(g_h),np.sin(b_h)+np.sin(g_h)])
g_dir = 0.5*np.array([np.cos(b_h)+np.cos(r_h),np.sin(b_h)+np.sin(r_h)])
b_dir = 0.5*np.array([np.cos(r_h)+np.cos(g_h),np.sin(r_h)+np.sin(g_h)])

r_dir /= r_dir[0]**2 + r_dir[1]**2
g_dir /= g_dir[0]**2 + g_dir[1]**2
b_dir /= b_dir[0]**2 + b_dir[1]**2

# These are coefficients to quickly test which component goes below zero first.
# Used like this in compute_max_saturation:
# if (-1.88170328f * a - 0.80936493f * b > 1) // Red component goes below zero first

print(r_dir)
print(g_dir)
print(b_dir)

r_hs,r_Ss = np.meshgrid(np.linspace(g_h,2*np.pi + b_h,200),np.linspace(0,1,200))

r_Rs = to_R(r_Ss, r_hs)

if PRINT_DIAGS:
    plt.imshow(np.sign(r_Rs), cmap='gray', vmin=0, vmax=1)
    plt.show()
    plt.figure()

g_hs,g_Ss = np.meshgrid(np.linspace(b_h,r_h,200),np.linspace(0,1,200))

g_Gs = to_G(g_Ss, g_hs)

if PRINT_DIAGS:
    plt.imshow(np.sign(g_Gs), cmap='gray', vmin=0, vmax=1)
    plt.show()
    plt.figure()

b_hs,b_Ss = np.meshgrid(np.linspace(r_h,g_h,200),np.linspace(0,1,200))

b_Bs = to_B(b_Ss, b_hs)

if PRINT_DIAGS:
    plt.imshow(np.sign(b_Bs), cmap='gray', vmin=0, vmax=1)
    plt.show()
    plt.figure()

# These are numerical fits to the edge of the chroma
# The resulting coefficient, x_R, x_G and x_B are used in compute_max_saturation, as values for k0

its = 1

h = np.linspace(g_h,2*np.pi + b_h,1000)
a = np.cos(h)
b = np.sin(h)

def e_R(x):
  S = x[0] + x[1]*a + x[2]*b + x[3]*a**2 + x[4]*a*b
  S = np.maximum(0, S)

  # optimize for solution that is easiest to solve with one step Haley's method
  f = to_R(S, h)
  f1 = to_R_dS(S, h)
  f2 = to_R_dS2(S, h)
  S_1 = S - f*f1/(f1**2 - f*f2/2)

  f_ = to_R(S_1, h)
  return np.average(f_**10) #+ f_[0]**2 + f_[-1]**2

x_R = scipy.optimize.minimize(e_R, np.array([1.19086277, 1.76576728, 0.59662641, 0.75515197, 0.56771245])).x

print(x_R)

S_R = x_R[0] + x_R[1]*a + x_R[2]*b + x_R[3]*a**2 + x_R[4]*a*b

plt.plot(S_R, 'r')
plt.figure()

plt.plot(to_R(S_R, h), 'r')
plt.figure()

S_R1 = S_R
for i in range(0,its):
  f = to_R(S_R1, h)
  f1 = to_R_dS(S_R1, h)
  f2 = to_R_dS2(S_R1, h)
  S_R1 = S_R1 - f*f1/(f1**2 - f*f2/(2))

  plt.plot(S_R1, 'r')

plt.figure()

plt.plot(to_R(S_R1, h), 'r')
plt.figure()

#####

h = np.linspace(b_h,r_h,1000)
a = np.cos(h)
b = np.sin(h)

def e_G(x):
  S = x[0] + x[1]*a + x[2]*b + x[3]*a**2 + x[4]*a*b
  S = np.maximum(0, S)

  # optimize for solution that is easiest to solve with one step Haley's method
  f = to_G(S, h)
  f1 = to_G_dS(S, h)
  f2 = to_G_dS2(S, h)
  S_1 = S - f*f1/(f1**2 - f*f2/2)

  f_ = to_G(S_1, h)
  return np.average(f_**10) #+ f_[0]**2 + f_[-1]**2

x_G = scipy.optimize.minimize(e_G, np.array([0.73956515, -0.45954404,  0.08285427,  0.12541073, -0.14503204])).x

print(x_G)

S_G = x_G[0] + x_G[1]*a + x_G[2]*b + x_G[3]*a**2 + x_G[4]*a*b

plt.plot(S_G, 'g')
plt.figure()

plt.plot(to_G(S_G, h), 'g')
plt.figure()

S_G1 = S_G
for i in range(0,its):
  f = to_G(S_G1, h)
  f1 = to_G_dS(S_G1, h)
  f2 = to_G_dS2(S_G1, h)
  S_G1 = S_G1 - f*f1/(f1**2 - f*f2/(2))

  plt.plot(S_G1, 'g')

plt.figure()

plt.plot(to_G(S_G1, h), 'g')
plt.figure()


#####

h = np.linspace(r_h,g_h,1000)
a = np.cos(h)
b = np.sin(h)

def e_B(x):
  S = x[0] + x[1]*a + x[2]*b + x[3]*a**2 + x[4]*a*b
  S = np.maximum(0, S)

  # optimize for solution that is easiest to solve with one step Haley's method
  f = to_B(S, h)
  f1 = to_B_dS(S, h)
  f2 = to_B_dS2(S, h)
  S_1 = S - f*f1/(f1**2 - f*f2/2)

  f_ = to_B(S_1, h)
  return np.average(f_**10) #+ f_[0]**2 + f_[-1]**2

x_B = scipy.optimize.minimize(e_B, np.array([1.35733652, -0.00915799, -1.1513021,  -0.50559606,  0.00692167])).x

print(x_B)

S_B = x_B[0] + x_B[1]*a + x_B[2]*b + x_B[3]*a**2 + x_B[4]*a*b

plt.plot(S_B, 'b')
plt.figure()

plt.plot(to_B(S_B, h), 'b')
plt.figure()

S_B1 = S_B
for i in range(0,its):
  f = to_B(S_B1, h)
  f1 = to_B_dS(S_B1, h)
  f2 = to_B_dS2(S_B1, h)
  S_B1 = S_B1 - f*f1/(f1**2 - f*f2/(2))

  plt.plot(S_B1, 'b')

plt.figure()

plt.plot(to_B(S_B1, h), 'b')
plt.figure()



"""Rec 2020 color class."""
from __future__ import annotations
from ..cat import WHITES
from .srgb import sRGB
import math
from .. import algebra as alg
from ..types import Vector

ALPHA = 1.09929682680944
BETA = 0.018053968510807
BETA45 = BETA * 4.5
ALPHAM1 = ALPHA - 1


def lin_2020(rgb: Vector) -> Vector:
    """
    Convert an array of rec-2020 RGB values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    https://en.wikipedia.org/wiki/Rec._2020#Transfer_characteristics
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA45:
            result.append(i / 4.5)
        else:
            result.append(math.copysign(alg.nth_root((abs_i + ALPHAM1) / ALPHA, 0.45), i))
    return result


def gam_2020(rgb: Vector) -> Vector:
    """
    Convert an array of linear-light rec-2020 RGB  in the range 0.0-1.0 to gamma corrected form.

    https://en.wikipedia.org/wiki/Rec._2020#Transfer_characteristics
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < BETA:
            result.append(4.5 * i)
        else:
            result.append(math.copysign(ALPHA * (abs_i ** 0.45) - ALPHAM1, i))
    return result


class Rec2020(sRGB):
    """Rec 2020 class."""

    BASE = "rec2020-linear"
    NAME = "rec2020"
    WHITE = WHITES['2deg']['D65']

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ from Rec. 2020."""

        return lin_2020(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ to Rec. 2020."""

        return gam_2020(coords)



"""
Slice diagrams.

Show a slice of a color space where we plot the relation between two channels holding one constant.

Channels to be plotted are given in the form `name:distance:offset`. For instance `a:1:-0.5` would plot the `a` channel
from -0.5 to 0.5.

The constant channel is given in the form `name:value` where value is the constant value for the channel.
"""
import matplotlib.pyplot as plt
import argparse
import sys
import os
import math
from shapely.geometry import Polygon
from shapely.geometry.base import dump_coords
from shapely.geometry.collection import GeometryCollection
from shapely.geometry.linestring import LineString

sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color
from coloraide import NaN  # noqa: E402
from coloraide.util import fmt_float  # noqa: E402
from coloraide.spaces import Cylindrical, LChish  # noqa: E402


def ignore_LCh_high_chroma_black(color):
    """
    Ignore LCh spaces that will render all lightness of zero with any chroma as the same color.

    Prevents a render of a single black line at lightness zero across all chroma.
    """

    names = color._space.names()
    return color[names[0]] == 0 and color[names[1]] != 0


def plot_slice(
    space,
    constants,
    channel1,
    channel2,
    gamut='srgb',
    resolution=500,
    dark=False,
    title="",
    subtitle='',
    polar=False,
    border=False,
    pointer=False
):
    """Plot a slice."""

    res = resolution
    if not dark:
        plt.style.use('seaborn-v0_8-darkgrid')
        default_color = 'black'
    else:
        plt.style.use('dark_background')
        default_color = 'white'

    figure = plt.figure()

    # Create a color object based on the specified space.
    c = Color(space, [])
    is_lchish = isinstance(c._space, LChish)

    # Parse the channel strings into actual values
    chan_constants = []
    for chan in constants.split(';'):
        n, v = [
            c if i == 0 else float(c if c != 'none' else 'nan') for i, c in enumerate(chan.split(':'), 0)
        ]
        n = c._space.CHANNEL_ALIASES.get(n, n)
        i = c._space.get_channel_index(n)
        chan_constants.append((n, v, i))

    name1, start1, end1 = [
        c if i == 0 else float(c) for i, c in enumerate(channel1.split(':'), 0)
    ]

    name2, start2, end2 = [
        c if i == 0 else float(c) for i, c in enumerate(channel2.split(':'), 0)
    ]

    # Get the actual indexes of the specified channels
    name1 = c._space.CHANNEL_ALIASES.get(name1, name1)
    index1 = c._space.get_channel_index(name1)
    name2 = c._space.CHANNEL_ALIASES.get(name2, name2)
    index2 = c._space.get_channel_index(name2)
    hue_index = -1

    kwargs = {}
    if polar and isinstance(c._space, Cylindrical):
        kwargs['projection'] = 'polar'
        index = c._space.hue_index()
        if index == index1:
            hue_index = index

    # Arrays for data points to plot
    c_map = []
    xaxis = []
    yaxis = []

    # Track minimum and maximum values
    c1_mn = float('inf')
    c1_mx = float('-inf')
    c2_mn = float('inf')
    c2_mx = float('-inf')

    # Track the edge of the graphed shape.
    # This is done from the perspective of the X and Y axis.
    # We will capture the min and max Y for a given X and
    # the min and max X for a given Y.
    # Both are imperfect, but taking both and find the
    # the intersection should yield a suitable edge.
    edge_map_x = {}
    edge_map_y = {}
    ex = []
    ey = []

    # Create initial coordinates for our 4 corners
    coords1 = [NaN] * (2 + len(chan_constants))
    for chan in chan_constants:
        coords1[chan[2]] = chan[1]
    coords1[index1] = start1
    coords1[index2] = start2

    coords2 = [NaN] * (2 + len(chan_constants))
    for chan in chan_constants:
        coords2[chan[2]] = chan[1]
    coords2[index1] = end1
    coords2[index2] = start2

    coords3 = [NaN] * (2 + len(chan_constants))
    for chan in chan_constants:
        coords3[chan[2]] = chan[1]
    coords3[index1] = start1
    coords3[index2] = end2

    coords4 = [NaN] * (2 + len(chan_constants))
    for chan in chan_constants:
        coords4[chan[2]] = chan[1]
    coords4[index1] = end1
    coords4[index2] = end2

    # Interpolate our min/max values along the x axis
    s1 = Color.steps([Color(space, coords1), Color(space, coords2)], steps=res, space=space, hue='specified')
    s2 = Color.steps([Color(space, coords3), Color(space, coords4)], steps=res, space=space, hue='specified')

    # Interpolate between each x axis color along the y axis
    for t1, t2 in zip(s1, s2):
        for r in Color.steps([t1, t2], steps=res, space=space, hue='specified'):
            # Only process colors within the specified gamut.
            if (
                r.in_gamut(gamut, tolerance=0) and
                (not pointer or r.in_pointer_gamut(tolerance=0)) and
                (not is_lchish or not ignore_LCh_high_chroma_black(r))
            ):
                c1 = r[index1]
                c2 = r[index2]

                if hue_index != -1:
                    c1 = math.radians(c1)

                # Get the absolute min and max value plotted
                if c1 < c1_mn:
                    c1_mn = c1
                if c1 > c1_mx:
                    c1_mx = c1

                if c2 < c2_mn:
                    c2_mn = c2
                if c2 > c2_mx:
                    c2_mx = c2

                if border:
                    # Create an edge map so we can draw an outline
                    # This should be all we need for polar.
                    if c1 not in edge_map_x:
                        mn = mx = c2
                    else:
                        mn, mx = edge_map_x[c1]
                        if c2 < mn:
                            mn = c2
                        elif c2 > mx:
                            mx = c2
                    edge_map_x[c1] = [mn, mx]

                    # If we aren't doing polar, we need both the X and Y perspective
                    if hue_index == -1:
                        if c2 not in edge_map_y:
                            mn = mx = c1
                        else:
                            mn, mx = edge_map_y[c2]
                            if c1 < mn:
                                mn = c1
                            elif c1 > mx:
                                mx = c1
                        edge_map_y[c2] = [mn, mx]

                # Save the points
                xaxis.append(c1)
                yaxis.append(c2)
                c_map.append(r.convert('srgb').to_string(hex=True))

    # Create axes
    ax = plt.axes(
        xlabel='{}: {} - {}'.format(name1, fmt_float(c1_mn, 5), fmt_float(c1_mx, 5)),
        ylabel='{}: {} - {}'.format(name2, fmt_float(c2_mn, 5), fmt_float(c2_mx, 5)),
        **kwargs
    )

    # Create titles
    if not title:
        title = "Plot of {} showing '{}' and '{}': {}".format(
            space, name1, name2, ' '.join(['{} = {}'.format(chan[0], fmt_float(chan[1], 5)) for chan in chan_constants])
        )

    plt.suptitle(title)
    if subtitle:
        if hue_index == -1:
            ax.set_title(subtitle, fontdict={'fontsize': 8})
        else:
            ax.set_title(subtitle, fontdict={'fontsize': 8}, pad=2)

    # Set aspect
    ax.set_aspect('auto' if hue_index == -1 else 'equal')
    figure.add_axes(ax)

    # Fill colors
    plt.scatter(
        xaxis,
        yaxis,
        marker="o",
        color=c_map,
        s=2
    )

    if border:
        # Create a border from the X axis perspective
        edge_x = []
        temp = []
        for p1, edges in edge_map_x.items():
            if hue_index == -1:
                # Min and max
                edge_x.append((p1, edges[0]))
                temp.append((p1, edges[1]))
            else:
                # Just need max for polar
                edge_x.append((p1, edges[1]))

        if hue_index == -1:
            # Combine the min/max values in one shape
            edge_x.extend(reversed(temp))
            # Close shape
            edge_x.append(edge_x[0])

        # Create a border from the Y axis perspective
        if hue_index == -1:
            edge_y = []
            temp.clear()

            for p1 in sorted(edge_map_y.keys()):
                edges = edge_map_y[p1]
                edge_y.append((edges[0], p1))
                temp.append((edges[1], p1))

            edge_y.extend(reversed(temp))
            edge_y.append(edge_y[0])

            # Get the intersection of X and Y to get the most accurate border
            poly = Polygon(edge_x).intersection(Polygon(edge_y))
            if isinstance(poly, GeometryCollection):
                for a in poly.geoms:
                    # Sometimes the intersection can have weird, unnecessary `LineString`.
                    # This occurs with very complex polygons. Just throw them away as
                    # they seem to be a bug.
                    if isinstance(a, LineString):
                        continue
                    ex, ey = zip(*dump_coords(a))
            else:
                ex, ey = zip(*dump_coords(poly))
        else:
            # Using polar, so just split the data by axis
            ex, ey = zip(*edge_x)

        # Plot the edge border
        plt.plot(ex, ey, color=default_color, marker="", linewidth=2, markersize=0, antialiased=True)


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='slice_diagrams', description='Plot a slice of a color space.')
    parser.add_argument('--space', '-s', help='Desired space.')
    parser.add_argument('--gamut', '-g', default="srgb", help='Gamut to evaluate the color in (default is sRGB).')
    parser.add_argument('--pointer', '-P', action='store_true', help="Restrict to Pointer gamut")
    parser.add_argument(
        '--constant', '-c', help="The channel(s) to hold constant and the value to use 'name:value;name2:value2'."
    )
    parser.add_argument('--xaxis', '-x', help="The channel to plot on X axis 'name:min:max'.")
    parser.add_argument('--yaxis', '-y', help="The channel to plot on Y axis 'name:min:max'.")
    parser.add_argument('--polar', '-p', action="store_true", help="Graph the cylindrical space in polar coordinates.")
    parser.add_argument('--resolution', '-r', default="800", help="How densely to render the figure.")
    parser.add_argument('--title', '-t', default='', help="Provide a title for the diagram.")
    parser.add_argument(
        '--sub-title', default='', help="Provide a subtitle, if none is provided, will show contant channel."
    )
    parser.add_argument('--no-border', '-b', action="store_true", help='Draw no border around the graphed content.')
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')

    args = parser.parse_args()

    plot_slice(
        args.space,
        args.constant,
        args.xaxis,
        args.yaxis,
        gamut=args.gamut,
        resolution=int(args.resolution),
        title=args.title,
        subtitle=args.sub_title,
        dark=args.dark,
        polar=args.polar,
        border=not args.no_border,
        pointer=args.pointer
    )

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""Parse utilities."""
from __future__ import annotations
import re
import math
from .. import algebra as alg
from ..types import Vector
from . import color_names
from ..channels import Channel, FLG_ANGLE
from typing import TYPE_CHECKING, Any
import functools

if TYPE_CHECKING:  # pragma: no cover
    from ..spaces import Space

RGB_CHANNEL_SCALE = 1.0 / 255.0
HUE_SCALE = 1.0 / 360.0
SCALE_PERCENT = 1 / 100.0
MAX_CHANNELS = 16

CONVERT_TURN = 360
CONVERT_GRAD = 90 / 100

RE_HEX = re.compile(r'(?i)(\#)((?:[a-f0-9]{6}(?:[a-f0-9]{2})?|[a-f0-9]{3}(?:[a-f0-9])?))\b')
RE_NAME = re.compile(r'(?i)\b([a-z]{3,})\b')
RE_IDENT = re.compile(r'(?i)(-{0,2}[a-z][-a-z0-9_]*)')
RE_SPACE = re.compile(r'(\s+)')
RE_LOOSE_SPACE = re.compile(r'(\s*)')
RE_CHANNEL = re.compile(r'(?i)((?:[+\-]?(?:[0-9]*\.)?[0-9]+(?:e[-+]?[0-9]+)?))(?:(%)|(deg|rad|turn|grad))?|(none)')
RE_FUNC_START = re.compile(r'(\()\s*')
RE_FUNC_END = re.compile(r'\s*(\))')
RE_COMMA = re.compile(r'\s*(,)\s*')
RE_SLASH = re.compile(r'\s*(/)\s*')
RE_CSS_FUNC = re.compile(r'\b(color|rgba?|hsla?|hwb|(?:ok)?lab|(?:ok)?lch)\b')


def norm_float(string: str) -> float:
    """Normalize a float value."""

    if string == "none":
        return alg.NaN
    return float(string)


def norm_hex_channel(string: str) -> float:
    """Normalize the hex string to a form we can handle."""

    return int(string, 16) * RGB_CHANNEL_SCALE


def norm_percent_channel(string: str, scale: float = 100, offset: float = 0.0) -> float:
    """Normalize percent channel."""

    if string == 'none':  # pragma: no cover
        return norm_float(string)
    elif string.endswith('%'):
        value = norm_float(string[:-1])
        return (value * scale * 0.01) - offset if scale != 100 else value
    else:  # pragma: no cover
        # Should only occur internally if we are doing something wrong.
        raise ValueError("Unexpected value '{}'".format(string))


def norm_color_channel(string: str, scale: float = 1, offset: float = 0.0) -> float:
    """Normalize percent/number channel."""

    if string.endswith('%'):
        return norm_percent_channel(string, scale, offset)
    else:
        return norm_float(string)


def norm_scaled_color_channel(string: str, scale: float = 1, offset: float = 0.0) -> float:
    """Normalize scaled percent/number channel."""

    if string.endswith('%'):
        return norm_percent_channel(string, scale, offset)
    else:
        value = norm_float(string)
        return (value * scale * 0.01) - offset if scale != 100 else value


def norm_rgb_channel(string: str, scale: float = 1) -> float:
    """Normalize RGB channel."""

    if string.endswith("%"):
        return norm_percent_channel(string, scale)
    else:
        return norm_float(string) * RGB_CHANNEL_SCALE


def norm_alpha_channel(string: str) -> float:
    """Normalize alpha channel."""

    if string.endswith("%"):
        value = norm_percent_channel(string, 1)
    else:
        value = norm_float(string)
    return alg.clamp(value, 0.0, 1.0)


def norm_angle_channel(angle: str) -> float:
    """Normalize angle units."""

    if angle.endswith('turn'):
        value = norm_float(angle[:-4]) * CONVERT_TURN
    elif angle.endswith('grad'):
        value = norm_float(angle[:-4]) * CONVERT_GRAD
    elif angle.endswith('rad'):
        value = math.degrees(norm_float(angle[:-3]))
    elif angle.endswith('deg'):
        value = norm_float(angle[:-3])
    else:
        value = norm_float(angle)
    return value


def parse_hex(color: str) -> tuple[Vector, float]:
    """Parse hexadecimal color."""

    length = len(color)
    if length in (7, 9):
        return (
            [
                norm_hex_channel(color[1:3]),
                norm_hex_channel(color[3:5]),
                norm_hex_channel(color[5:7])
            ],
            norm_hex_channel(color[7:9]) if length == 9 else 1.0
        )
    else:
        return (
            [
                norm_hex_channel(color[1] * 2),
                norm_hex_channel(color[2] * 2),
                norm_hex_channel(color[3] * 2)
            ],
            norm_hex_channel(color[4] * 2) if length == 5 else 1.0
        )


def parse_rgb_channels(color: list[str], boundry: tuple[Channel, ...]) -> tuple[Vector, float]:
    """Parse CSS RGB format."""

    channels = []
    alpha = 1.0
    for i, c in enumerate(color, 0):
        c = c.lower()
        if i <= 2:
            channels.append(norm_rgb_channel(c, boundry[i].high))
        elif i == 3:
            alpha = norm_alpha_channel(c)
    return channels, alpha


def parse_channels(color: list[str], boundry: tuple[Channel, ...], scaled: bool = False) -> tuple[Vector, float]:
    """Parse CSS channel format."""

    channels = []
    alpha = 1.0
    length = len(boundry)
    for i, c in enumerate(color, 0):
        c = c.lower()
        if i < length:
            bound = boundry[i]
            if bound.flags & FLG_ANGLE:
                channels.append(norm_angle_channel(c))
            elif scaled:
                channels.append(norm_scaled_color_channel(c, bound.high))
            else:
                channels.append(norm_color_channel(c, bound.high))
        elif i == length:
            alpha = norm_alpha_channel(c)
    return channels, alpha


def parse_color(tokens: dict[str, Any], space: Space) -> tuple[Vector, float] | None:
    """Parse the color function."""

    # Iterate the spaces and see if we find the color serialization identifier
    num_channels = len(space.CHANNELS)
    values = len(tokens['func']['values'])

    if tokens['func']['slash']:
        values -= 1

    if values != num_channels:
        return None

    alpha = norm_alpha_channel(tokens['func']['values'][-1]['value']) if tokens['func']['slash'] else 1.0

    channels = []
    properties = space.CHANNELS
    for i in range(num_channels):
        c = tokens['func']['values'][i]['value']
        channel = properties[i]
        channels.append(norm_color_channel(c.lower(), channel.span, channel.offset))
    return (channels, alpha)


def validate_color(tokens: dict[str, Any]) -> bool:
    """Validate the color function syntax."""

    return not any(v['type'] == 'degree' for v in tokens['func']['values'])


def validate_srgb(tokens: dict[str, Any]) -> bool:
    """Validate the RGB color functions."""

    length = len(tokens['func']['values'])
    delimiter = tokens['func']['delimiter']

    if length < 3 or length > 4:
        return False
    vtype = tokens['func']['values'][0]['type']
    if delimiter == 'comma':
        if vtype in ('none', 'degree') or not all(v['type'] == vtype for v in tokens['func']['values'][:3]):
            return False
        if length == 4 and tokens['func']['values'][3]['type'] in ('none', 'degree'):
            return False
    else:
        if any(v['type'] == 'degree' for v in tokens['func']['values']):
            return False
        if length == 4 and not tokens['func']['slash']:
            return False
    return True


def validate_cylindrical_srgb(tokens: dict[str, Any]) -> bool:
    """Validate cylindrical sRGB."""

    length = len(tokens['func']['values'])
    delimiter = tokens['func']['delimiter']
    func_name = tokens['func']['name']

    if length < 3 or length > 4:
        return False

    if func_name == 'hwb' and delimiter == 'comma':
        return False

    if delimiter == 'comma':
        if tokens['func']['values'][0]['type'] not in ('degree', 'number'):
            return False
        if not all(v['type'] == 'percent' for v in tokens['func']['values'][1:3]):
            return False
        if length == 4 and tokens['func']['values'][3]['type'] in ('none', 'degree'):
            return False
    else:
        if tokens['func']['values'][0]['type'] == 'percent':
            return False
        if any(v['type'] == 'degree' for v in tokens['func']['values'][1:]):
            return False

        if length == 4 and not tokens['func']['slash']:
            return False
    return True


def validate_lab(tokens: dict[str, Any]) -> bool:
    """Validate CSS Lab variant color spaces."""

    length = len(tokens['func']['values'])
    delimiter = tokens['func']['delimiter']

    if delimiter == 'comma':
        return False

    if length < 3 or length > 4:
        return False

    if any(v['type'] == 'degree' for v in tokens['func']['values'][:]):
        return False

    if length == 4 and not tokens['func']['slash']:
        return False

    return True


def validate_lch(tokens: dict[str, Any]) -> bool:
    """Validate CSS LCh variant color spaces."""

    length = len(tokens['func']['values'])
    delimiter = tokens['func']['delimiter']

    if delimiter == 'comma':
        return False

    if length < 3 or length > 4:
        return False

    if tokens['func']['values'][2]['type'] == 'percent':
        return False

    if any(v['type'] == 'degree' for v in tokens['func']['values'][0:2]):
        return False

    if length == 4 and (not tokens['func']['slash'] or tokens['func']['values'][3]['type'] == 'degree'):
        return False

    return True


@functools.lru_cache(maxsize=1)
def tokenize_css(css: str, start: int = 0) -> dict[str, Any]:
    """Tokenize the CSS string."""

    tokens = {}  # type: dict[str, Any]
    # `mypy` will get confused, just set to Any
    m = RE_HEX.match(css, start)  # type: Any
    if m:
        tokens['hex'] = {
            'start': m.group(1),
            'value': m.group(0)
        }
        tokens['id'] = 'srgb'
        tokens['end'] = m.end()
        return tokens
    m = RE_NAME.match(css, start)
    if m:
        # Is hex?
        if color_names.has_name(m.group(1)):
            tokens['name'] = {'color': m.group(1)}
            tokens['id'] = 'srgb'
            tokens['end'] = m.end()
            return tokens

        # Is CSS function
        func_name = m.group(0).lower()
        m2 = RE_CSS_FUNC.match(func_name)
        if not m2:
            return {}

        # Has function body?
        tokens['func'] = {'name': func_name, 'values': [], 'delimiter': ''}
        m = RE_FUNC_START.match(css, m.end())
        if not m:  # pragma: no cover
            return {}

        # If a color function, does it have an identifier?
        delimiter = None
        if func_name == 'color':
            m2 = RE_IDENT.match(css, m.end())
            if not m2:
                return {}
            delimiter = 'space'
            tokens['func']['delimiter'] = delimiter
            tokens['id'] = m2.group(1)
            m = RE_SPACE.match(css, m2.end())
            if not m:
                return {}

        # Gather function channel values, up to 16.
        slash = False
        for _ in range(MAX_CHANNELS):
            # Get channel value
            m2 = RE_CHANNEL.match(css, m.end())
            if not m2:
                if slash:
                    return {}
                break
            m = m2
            if m.group(2):
                tokens['func']['values'].append({'type': 'percent', 'value': m.group(0)})
            elif m.group(3):
                tokens['func']['values'].append({'type': 'degree', 'value': m.group(0)})
            elif m.group(4):
                tokens['func']['values'].append({'type': 'none', 'value': m.group(0)})
            else:
                tokens['func']['values'].append({'type': 'number', 'value': m.group(0)})

            if slash:
                break

            # Get delimiter type
            if delimiter is None:
                m2 = RE_COMMA.match(css, m.end(0))
                if not m2:
                    m2 = RE_LOOSE_SPACE.match(css, m.end(0))
                    if m2:
                        delimiter = 'space'
                        tokens['func']['delimiter'] = delimiter
                else:
                    delimiter = 'comma'
                    tokens['func']['delimiter'] = delimiter

            # Find comma
            elif delimiter == 'comma':
                m2 = RE_COMMA.match(css, m.end(0))
                if not m2:
                    break

            # Find space/slash
            else:
                m2 = RE_SLASH.match(css, m.end(0))
                if m2:
                    slash = True
                else:
                    m2 = RE_LOOSE_SPACE.match(css, m.end(0))

            m = m2

        tokens['func']['slash'] = slash

        # Get function end
        m = RE_FUNC_END.match(css, m.end())
        if not m:
            return {}

        # Do basic validation on the supported color functions
        tokens['end'] = m.end()
        if func_name == 'color' and not validate_color(tokens):
            return {}

        elif func_name.startswith('rgb'):
            tokens['id'] = 'srgb'
            if not validate_srgb(tokens):
                return {}

        elif func_name in ('hsl', 'hsla', 'hwb'):
            tokens['id'] = '--hwb' if func_name == 'hwb' else '--hsl'

            if not validate_cylindrical_srgb(tokens):
                return {}

        elif func_name in ('lab', 'oklab'):
            tokens['id'] = '--' + func_name

            if not validate_lab(tokens):
                return {}

        elif func_name in ('oklch', 'lch'):
            tokens['id'] = '--' + func_name

            if not validate_lch(tokens):
                return {}

    return tokens


def parse_css(
    cspace: Space,
    string: str,
    start: int = 0,
    fullmatch: bool = True,
    color: bool = False
) -> tuple[tuple[Vector, float], int] | None:
    """Match a CSS color string."""

    target = cspace.SERIALIZE
    if not target:
        target = (cspace.NAME,)

    tokens = tokenize_css(string, start=start)

    # Could we parse it?
    if not tokens:
        return None

    # Does the space match the target space?
    if tokens['id'] not in target:
        return None

    # Did we get a full match if requested?
    if fullmatch and tokens['end'] < len(string):
        return None

    # Scale and convert the channel data according to the CSS space rules
    end = tokens['end']
    if 'func' in tokens and tokens['func']['name'] == 'color':
        if color is False:
            return None

        result = parse_color(tokens, cspace)
        if result is None:
            return result
        return result, end

    elif tokens['id'] == 'srgb':
        if 'hex' in tokens:
            return parse_hex(tokens['hex']['value']), end
        elif 'name' in tokens:
            values = color_names.from_name(tokens['name']['color'])
            return (values[:-1], values[-1]), end  # type: ignore[index]
        else:
            return parse_rgb_channels([v['value'] for v in tokens['func']['values']], cspace.CHANNELS), end
    elif tokens['id'] in ('--hsl', '--hwb'):
        return parse_channels([v['value'] for v in tokens['func']['values']], cspace.CHANNELS, scaled=True), end
    else:
        return parse_channels([v['value'] for v in tokens['func']['values']], cspace.CHANNELS), end



"""Provides a plugin system for filtering colors."""
from __future__ import annotations
from abc import ABCMeta, abstractmethod
from ..types import Plugin
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class Filter(Plugin, metaclass=ABCMeta):
    """Filter a color."""

    NAME = ''
    DEFAULT_SPACE = 'srgb-linear'
    ALLOWED_SPACES = ('srgb-linear',)  # type: tuple[str, ...]

    @abstractmethod
    def filter(self, color: Color, amount: float | None, **kwargs: Any) -> None:  # noqa: A003
        """Filter the given color."""


def filters(
    color: Color,
    name: str,
    amount: float | None = None,
    space: str | None = None,
    out_space: str | None = None,
    in_place: bool = False,
    **kwargs: Any
) -> Color:
    """Filter."""

    f = color.FILTER_MAP.get(name)
    if not f:
        raise ValueError("'{}' filter is not supported".format(name))

    if space is None:
        space = f.DEFAULT_SPACE

    if space not in f.ALLOWED_SPACES:
        raise ValueError(
            "The '{}' only supports filtering in the {} spaces, not '{}'".format(name, str(f.ALLOWED_SPACES), space)
        )

    if out_space is None:
        out_space = space

    c = color.convert(space, in_place=in_place, norm=False).normalize()
    f.filter(c, amount, **kwargs)
    return c.convert(out_space, in_place=True)



"""Plot easing functions."""
import argparse
import sys
import os
import re
import matplotlib.pyplot as plt

sys.path.insert(0, os.getcwd())

from coloraide import easing  # noqa: E402

RE_BEZIER = re.compile(
    r'''(?x)
    ^(cubic_bezier)\((
        (?:\s*[+\-]?(?:[0-9]*\.)?[0-9]+(?:e[-+]?[0-9]*)?\s*,){3}
        \s*[+\-]?(?:[0-9]*\.)?[0-9]+(?:e[-+]?[0-9]*)?\s*
    )\)$
    '''
)


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='easing_diagrams', description='Plot easings.')
    parser.add_argument('--easing', '-s', default='linear', help="Space to interpolate in.")
    parser.add_argument('--extrapolate', '-e', action='store_true', help='Extrapolate values.')
    parser.add_argument('--title', '-T', default='', help="Provide a title for the diagram.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')
    args = parser.parse_args()

    m = RE_BEZIER.match(args.easing.strip())
    if m is not None:
        function = m.group(0).strip()
        ease = easing.cubic_bezier(*[float(v.strip()) for v in m.group(2).split(',')])
    else:
        function = args.easing.strip()
        ease = getattr(easing, function)

    xs = []
    ys = []
    if not args.extrapolate:
        offset, factor = 0, 100
    else:
        offset, factor = 25, 50
    for r in range(101):
        x = (r - offset) / factor
        xs.append(x)
        ys.append(ease(x))

    plt.style.use('seaborn-v0_8-darkgrid')
    default_color = 'black'

    figure = plt.figure()

    # Create axes
    ax = plt.axes(
        xlabel='Time',
        ylabel='Progression'
    )

    ax.set_aspect('equal')

    # Create titles
    title = args.title
    if not title:
        title = function
    ax.set_title(title)

    figure.add_axes(ax)

    p0 = [0, 0]
    p3 = [1, 1]
    if function == 'linear':
        p1 = [0.0, 0.0]
        p2 = [1.0, 1.0]
    else:
        p1 = ease.keywords['p1']
        p2 = ease.keywords['p2']

    plt.plot(
        xs,
        ys,
        color=default_color,
        marker="",
        linewidth=1.5,
        markersize=2,
        antialiased=True
    )

    plt.plot(
        [p0[0], p1[0]],
        [p0[1], p1[1]],
        '--',
        color='teal',
        marker="",
        linewidth=1.5,
        markersize=2,
        antialiased=True
    )

    plt.plot(
        [p2[0], p3[0]],
        [p2[1], p3[1]],
        '--',
        color='teal',
        marker="",
        linewidth=1.5,
        markersize=2,
        antialiased=True
    )

    plt.scatter(
        [p0[0], p3[0]],
        [p0[1], p3[1]],
        marker='o',
        color='black',
        zorder=2
    )

    plt.scatter(
        [p1[0], p2[0]],
        [p1[1], p2[1]],
        marker='o',
        color='teal',
        zorder=2
    )

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""Modify a picture with a given filter."""
from functools import lru_cache
from PIL import Image
import time
import argparse
import sys
import os

# We want to load ColorAide from the working directory to pick up the version under development
sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color


def printt(t):
    """Print time."""

    print('Completed in: ', end='')
    s = t // 1e+9
    m = t // 1e+6
    u = t // 1000
    if s:
        s = t / 1e+9
        h = m = 0
        m = s // 60
        if m:
            s -= m * 60
            h = m // 60
            if h:
                m -= h * 60
        if h:
            print('{} hours '.format(h), end='')
        if m:
            print('{} minutes '.format(m), end='')
        print('{} sec'.format(s))
    elif m:
        print('{} msec'.format(t / 1e+6))
    elif u:
        print('{} usec'.format(t / 1000))
    else:
        print('{} nsec'.format(t))


@lru_cache(maxsize=1024 * 1024)
def apply_filter(name, amount, space, method, p, fit):
    """Apply filter."""

    has_alpha = len(p) > 3
    color = Color('srgb', [x / 255 for x in p[:3]], p[3] / 255 if has_alpha else 1)
    if method is not None:
        # This is a CVD filter that allows specifying the method
        color.filter(name, amount, space=space, in_place=True, method=method, out_space='srgb')
    else:
        # General filter.
        color.filter(name, amount, space=space, in_place=True, out_space='srgb')
    # Fit the color back into the color gamut and return the results
    return tuple(int(x * 255) for x in color.fit(method=fit)[:4 if has_alpha else -1])


def process_image(img, output, name, amount, space, cvd_approach, fit):
    """Process the image applying the requested filter."""

    with Image.open(img) as im:

        # Make sure we are writing in transparent mode for the opacity filter
        if im.format == 'PNG' and name == 'opacity':
            if im.mode not in ('RGBA',):
                im = im.convert('RGBA')

        pixels = im.load()
        start = time.perf_counter_ns()
        total = im.size[0] * im.size[1]
        factor = 100 / total
        i = j = 0
        print('Pixels: {}'.format(total))
        print('> 0%', end='\r')
        for i in range(im.size[0]):
            for j in range(im.size[1]):
                pixels[i, j] = apply_filter(name, amount, space, cvd_approach, pixels[i, j], fit)
            print('> {}%'.format(int((i * j) * factor)), end="\r")
        print('> 100%')
        t = time.perf_counter_ns() - start
        printt(t)
        im.save(output)


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='filter_img', description='Apply filter to an image.')
    parser.add_argument('--input', '-i', help='Input image.')
    parser.add_argument('--output', '-o', help='Output name and location.')
    parser.add_argument('--filter', '-f', help='The filter to use.')
    parser.add_argument('--amount', '-a', type=float, help='Amount to filter the image.')
    parser.add_argument('--cvd-approach', '-c', help='CVD approach to use.')
    parser.add_argument('--space', '-s', default='srgb-linear', help='Color space to filter in.')
    parser.add_argument('--gamut-map', '-g', default="clip", help="Specify GMA method to use (default simple clipping)")
    args = parser.parse_args()

    process_image(
        args.input,
        args.output,
        args.filter,
        args.amount,
        args.space,
        args.cvd_approach,
        args.gamut_map
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Natural B-Spline interpolation.

https://www.math.ucla.edu/~baker/149.1.02w/handouts/dd_splines.pdf.
"""
from __future__ import annotations
from .. interpolate import Interpolate, Interpolator
from .bspline import InterpolatorBSpline
from .. import algebra as alg
from ..types import Vector
from typing import List, Sequence, Any, Mapping, Callable, Dict, Type, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class InterpolatorNaturalBSpline(InterpolatorBSpline):
    """Natural B-spline class."""

    def setup(self) -> None:
        """
        Using B-spline as the base create a natural spline that also passes through the control points.

        Using the color points as `S0...Sn`, calculate `B0...Bn`, such that interpolation will
        pass through `S0...Sn`.

        When given 2 data points, the operation will be linear, so there is nothing to do.
        """

        # Use the same logic as normal B-spline for handling undefined values and applying premultiplication
        self.spline = alg.bspline
        self.handle_undefined()
        alg.naturalize_bspline_controls(self.coordinates)
        self.adjust_endpoints()


class NaturalBSpline(Interpolate):
    """Natural B-spline interpolation plugin."""

    NAME = "natural"

    def interpolator(
        self,
        coordinates: List[Vector],
        channel_names: Sequence[str],
        create: Type[Color],
        easings: List[Callable[..., float] | None],
        stops: Dict[int, float],
        space: str,
        out_space: str,
        progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
        premultiplied: bool,
        extrapolate: bool = False,
        domain: list[float] | None = None,
        padding: float | tuple[float, float] | None = None,
        **kwargs: Any
    ) -> Interpolator:
        """Return the natural B-spline interpolator."""

        return InterpolatorNaturalBSpline(
            coordinates,
            channel_names,
            create,
            easings,
            stops,
            space,
            out_space,
            progress,
            premultiplied,
            extrapolate,
            domain,
            padding,
            **kwargs
        )



"""
Catmull-Rom interpolation.

http://www2.cs.uregina.ca/~anima/408/Notes/Interpolation/Parameterized-Curves-Summary.htm
"""
from __future__ import annotations
from .bspline import InterpolatorBSpline
from ..interpolate import Interpolator, Interpolate
from .. import algebra as alg
from ..types import Vector
from typing import Callable, Mapping, Sequence, Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class InterpolatorCatmullRom(InterpolatorBSpline):
    """Interpolate with Catmull-Rom spline."""

    def setup(self) -> None:
        """Setup."""

        super().setup()
        self.spline = alg.catrom


class CatmullRom(Interpolate):
    """Catmull-Rom interpolation plugin."""

    NAME = "catrom"

    def interpolator(
        self,
        coordinates: list[Vector],
        channel_names: Sequence[str],
        create: type[Color],
        easings: list[Callable[..., float] | None],
        stops: dict[int, float],
        space: str,
        out_space: str,
        progress: Mapping[str, Callable[..., float]] | Callable[..., float] | None,
        premultiplied: bool,
        extrapolate: bool = False,
        domain: list[float] | None = None,
        padding: float | tuple[float, float] | None = None,
        **kwargs: Any
    ) -> Interpolator:
        """Return the Catmull-Rom interpolator."""

        return InterpolatorCatmullRom(
            coordinates,
            channel_names,
            create,
            easings,
            stops,
            space,
            out_space,
            progress,
            premultiplied,
            extrapolate,
            domain,
            padding,
            **kwargs
        )



"""Test Interpolation."""
import unittest
from coloraide.everything import ColorAll as Color
from coloraide import NaN, stop, hint, ease_in
from . import util
import pytest


@pytest.mark.skipif(True, reason='Experimental')
class TestCarryFoward(util.ColorAssertsPyTest):
    """Test carry forward."""

    CASES = [
        [
            'oklab',
            4,
            ['lch(none 30 270)', 'lab(none -2 12)', 'hsl(30 80 none)', 'oklab(0.5 0.3 -0.1)'],
            ('l', 0.5)
        ],
        [
            'oklch',
            5,
            [
                'lch(40 none 270)', 'hsl(30 none 75)', 'color(--hsv 300 none 0.5)',
                'color(--hsi 90 none 0.2)', 'color(--oklch 0.5 0.2 30)'],
            ('c', 0.2)
        ],
        [
            'hsl',
            5,
            [
                'lch(40 none 270)', 'color(--hsv 300 none 0.5)', 'color(--hsi 90 none 0.2)',
                'color(--oklch 0.5 none 30)', 'hsl(30 30 75)'],
            ('s', 0.3)
        ],
        [
            'hsv',
            5,
            [
                'lch(40 none 270)', 'hsl(30 none 75)', 'color(--hsi 90 none 0.2)',
                'color(--oklch 0.5 none 30)', 'color(--hsv 300 0.3 0.5)'],
            ('s', 0.3)
        ],
        [
            'hsl',
            6,
            [
                'lch(40 10 none)', 'color(--hsv none 0.1 0.5)', 'hwb(none 30 75)',
                'color(--hsi none 0.1 0.2)', 'color(--oklch 0.5 0.1 none)', 'hsl(30 30 75)'],
            ('h', 30)
        ],
        [
            'hwb',
            6,
            [
                'lch(40 10 none)', 'color(--hsv none 0.1 0.5)', 'hsl(none 30 75)',
                'color(--hsi none 0.1 0.2)', 'color(--oklch 0.5 0.1 none)', 'hwb(30 30 75)'],
            ('h', 30)
        ],
        [
            'oklab',
            3,
            [
                'lch(40 10 270 / none)', 'rgb(220 0 47 / none)', 'hsl(30 30 75 / 0.5)',
            ],
            ('alpha', 0.5)
        ],
        [
            'srgb',
            3,
            [
                'color(xyz-d65 0.24 0.34 none)', 'color(display-p3 0 1 none)', 'rgb(30 30 75)'
            ],
            ('blue', 75 / 255)
        ]
    ]

    @pytest.mark.parametrize('space, steps, colors, cmp', CASES)
    def test_round_trip(self, space, steps, colors, cmp):
        """Test round trip."""

        results = Color.steps(colors, steps=steps, space=space, method='monotone', _carryforward=True)
        assert not all(abs(r[cmp[0]] - cmp[1]) < 1e-6 for r in results), "{} != {} : {}".format(cmp[0], cmp[1], results)


class TestInterpolation(util.ColorAsserts, unittest.TestCase):
    """Test interpolation."""

    def test_inputs_unmodified(self):
        """Test that interpolation inputs are unmodified."""

        c1 = Color('red')
        c2 = Color('blue')
        _ = Color.interpolate([c1, c2], space='oklab')
        self.assertEqual(c1.to_string(), 'rgb(255 0 0)')
        self.assertEqual(c2.to_string(), 'rgb(0 0 255)')

    def test_out_space_cylindrical(self):
        """Test out space with cylindrical color space."""

        i = Color.interpolate(['transparent', 'hsl(30deg 30% 40%)'], steps=100, space='hsl', out_space='srgb')
        self.assertColorEqual(i(0.5), Color('srgb', [0.52, 0.4, 0.28], 0.5))

    def test_domain(self):
        """Test that domains work."""

        i = Color.interpolate(
            ['blue', 'green', 'yellow', 'orange', 'red'],
            domain=[-32, 32, 60, 85, 95],
        )

        self.assertColorEqual(i(-32), Color('oklab(0.45201 -0.03246 -0.31153)'))
        self.assertColorEqual(i(35), Color('oklab(0.56778 -0.13292 0.11741)'))
        self.assertColorEqual(i(60), Color('oklab(0.96798 -0.07137 0.19857)'))
        self.assertColorEqual(i(79), Color('oklab(0.83476 0.0259 0.17031)'))
        self.assertColorEqual(i(95), Color('oklab(0.62796 0.22486 0.12585)'))

    def test_domain_extrapolation(self):
        """Test extrapolation with custom domain."""

        i = Color.interpolate(['red', 'blue'], extrapolate=True, domain=[-25, 25])
        self.assertColorEqual(i(-30), Color('oklab(0.64555 0.2506 0.16958)'))
        self.assertColorEqual(i(30), Color('oklab(0.43442 -0.05819 -0.35527)'))

    def test_domain_of_one(self):
        """
        Test domain of one.

        This is pointless, but we won't break because of it.
        """

        i = Color.interpolate(
            ['blue', 'green', 'yellow', 'orange', 'red'],
            domain=[1]
        )

        self.assertColorEqual(i(-1), Color('oklab(0.45201 -0.03246 -0.31153)'))
        self.assertColorEqual(i(0), Color('oklab(0.45201 -0.03246 -0.31153)'))
        self.assertColorEqual(i(1), Color('oklab(0.45201 -0.03246 -0.31153)'))
        self.assertColorEqual(i(2), Color('oklab(0.62796 0.22486 0.12585)'))

    def test_domain_in_step(self):
        """Test domains work in steps."""

        steps = Color.steps(
            ['blue', 'green', 'yellow', 'orange', 'red'],
            steps=11,
            domain=[-32, 32, 60, 85, 95]
        )

        self.assertColorEqual(steps[0], Color('oklab(0.45201 -0.03246 -0.31153)'))
        self.assertColorEqual(steps[3], Color('oklab(0.49234 -0.09666 -0.06197)'))
        self.assertColorEqual(steps[5], Color('oklab(0.51922 -0.13946 0.1044)'))
        self.assertColorEqual(steps[7], Color('oklab(0.91836 -0.079 0.18851)'))
        self.assertColorEqual(steps[10], Color('oklab(0.62796 0.22486 0.12585)'))

    def test_domain_mix(self):
        """Test domains in mix."""

        self.assertColorEqual(
            Color('red').mix('blue', 0.75, domain=[0.0, 0.75, 1.0]),
            Color('red').mix('blue', 0.5)
        )

    def test_mix(self):
        """Test interpolation via mixing."""

        self.assertColorEqual(Color('red').mix('blue', 1), Color("color(--oklab 0.45201 -0.03246 -0.31153 / 1)"))
        self.assertColorEqual(Color('red').mix('blue', 0.75), Color("color(--oklab 0.496 0.03187 -0.20218 / 1)"))
        self.assertColorEqual(Color('red').mix('blue'), Color("color(--oklab 0.53998 0.0962 -0.09284 / 1)"))
        self.assertColorEqual(Color('red').mix('blue', 0.25), Color("color(--oklab 0.58397 0.16053 0.0165 / 1)"))
        self.assertColorEqual(Color('red').mix('blue', 0.0), Color("color(--oklab 0.62796 0.22486 0.12585 / 1)"))

    def test_mix_dict(self):
        """Test mixing with a mapping."""

        c1 = Color('blue')
        self.assertEqual(
            c1.mix("yellow"),
            c1.mix({"space": "srgb", "coords": [1, 1, 0]})
        )

    def test_bad_mix_input(self):
        """Test bad mix input."""

        with self.assertRaises(TypeError):
            Color('red').mix(1)

    def test_mix_input_piecewise(self):
        """Test mix with piecewise."""

        with self.assertRaises(TypeError):
            Color('red').mix(stop('blue', 0.0), 0.5, space="srgb")

    def test_mix_space(self):
        """Test color mix in different space."""

        self.assertColorEqual(Color('red').mix('blue', 1, space="srgb"), Color("srgb", [0, 0, 1]))
        self.assertColorEqual(Color('red').mix('blue', 0.75, space="srgb"), Color("srgb", [0.25, 0, 0.75]))
        self.assertColorEqual(Color('red').mix('blue', space="srgb"), Color("srgb", [0.5, 0, 0.5]))
        self.assertColorEqual(Color('red').mix('blue', 0.25, space="srgb"), Color("srgb", [0.75, 0, 0.25]))
        self.assertColorEqual(Color('red').mix('blue', 0.0, space="srgb"), Color("srgb", [1, 0, 0]))

    def test_mix_out_space(self):
        """Test interpolation."""

        self.assertColorEqual(
            Color('red').mix('blue', 1, space="lab", out_space="lab"),
            Color("lab(29.568% 68.287 -112.03)")
        )
        self.assertColorEqual(
            Color('red').mix('blue', 0.75, space="lab", out_space="lab"),
            Color("lab(35.749% 71.417 -66.55)")
        )
        self.assertColorEqual(
            Color('red').mix('blue', space="lab", out_space="lab"),
            Color("lab(41.929% 74.546 -21.069)")
        )
        self.assertColorEqual(
            Color('red').mix('blue', 0.25, space="lab", out_space="lab"),
            Color("lab(48.11% 77.676 24.411)")
        )
        self.assertColorEqual(
            Color('red').mix('blue', 0.0, space="lab", out_space="lab"),
            Color("lab(54.291% 80.805 69.891)")
        )

    def test_mix_alpha(self):
        """Test mixing alpha."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 0.75)').mix('color(srgb 0 0 1 / 0.25)', space="srgb"),
            Color('rgb(191.25 0 63.75 / 0.5)')
        )

    def test_mix_premultiplied_alpha(self):
        """Test premultiplied alpha."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0 / 0.75)').mix('color(srgb 0 0 1 / 0.25)', premultiplied=True, space="srgb"),
            Color('rgb(191.25 0 63.75 / 0.5)')
        )

    def test_mix_premultiplied_no_alpha(self):
        """Test premultiplied alpha."""

        self.assertColorEqual(
            Color('color(srgb 1 0 0)').mix('color(srgb 0 0 1)', premultiplied=True, space="srgb"),
            Color('color(srgb 1 0 0)').mix('color(srgb 0 0 1)', space="srgb")
        )

    def test_mix_premultiplied_cylindrical(self):
        """Test premultiplication in a cylindrical space."""

        self.assertColorEqual(
            Color('color(--hsl 20 30% 75% / 0.5)').mix(
                'color(--hsl 20 60% 10% / 0.75)', premultiplied=True, space="hsl"
            ),
            Color('hsl(20 48% 36% / 0.625)')
        )

    def test_mix_in_place(self):
        """Test mix in place."""

        color = Color('red')
        color2 = color.mix('blue', space="srgb")
        self.assertIsNot(color, color2)
        self.assertColorEqual(color2, Color("srgb", [0.5, 0, 0.5]))
        color = Color('red')
        color2 = color.mix('blue', space="srgb", in_place=True)
        self.assertIs(color, color2)
        self.assertColorEqual(color, Color("srgb", [0.5, 0, 0.5]))

    def test_mix_nan(self):
        """Test mixing with NaN."""

        c1 = Color("srgb", [NaN, 1, 1])
        c2 = Color("srgb", [0.75, 0, 0])
        self.assertColorEqual(c1.mix(c2, space="srgb"), Color("srgb", [0.75, 0.5, 0.5]))
        c1 = Color("srgb", [0.25, 1, 1])
        c2 = Color("srgb", [NaN, 0, 0])
        self.assertColorEqual(c1.mix(c2, space="srgb"), Color("srgb", [0.25, 0.5, 0.5]))
        c1 = Color("srgb", [NaN, 1, 1])
        c2 = Color("srgb", [NaN, 0, 0])
        self.assertColorEqual(c1.mix(c2, space="srgb"), Color("srgb", [0, 0.5, 0.5]))

    def test_mix_mask(self):
        """Test mix adjust method."""

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(c1.mix(c2.mask("red"), space="srgb"), Color("srgb", [0.25, 0.5, 0.5]))

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(c1.mask("red").mix(c2, space="srgb"), Color("srgb", [0.75, 0.5, 0.5]))

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(c1.mask("red").mix(c2.mask("red"), space="srgb"), Color("srgb", [0.0, 0.5, 0.5]))

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(c1.mix(c2.mask(["red", "green"]), space="srgb"), Color("srgb", [0.25, 1, 0.5]))

    def test_mix_mask_invert(self):
        """Test mix adjust method."""

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(
            c1.mix(c2.mask(["green", "blue"], invert=True), space="srgb"),
            Color("srgb", [0.25, 0.5, 0.5])
        )

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(
            c1.mask(["green", "blue"], invert=True).mix(c2, space="srgb"),
            Color("srgb", [0.75, 0.5, 0.5])
        )

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(
            c1.mask(["green", "blue", "alpha"], invert=True).mix(
                c2.mask(["green", "blue", "alpha"], invert=True),
                space="srgb"
            ),
            Color("srgb", [0.0, 0.5, 0.5])
        )

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(
            c1.mix(c2.mask("blue", invert=True), space="srgb"),
            Color("srgb", [0.25, 1, 0.5])
        )

    def test_mix_hue_adjust(self):
        """Test hue adjusting."""

        c1 = Color('rebeccapurple')
        c2 = Color('lch(85% 100 805)')
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.25, hue="shorter", space="lch"),
            Color("lch(32.393 61.244 342.89)")
        )
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.25, hue="longer", space="lch"),
            Color("lch(32.393 61.244 252.89)")
        )
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.25, hue="increasing", space="lch"),
            Color("lch(32.393 61.244 342.89)")
        )
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.25, hue="decreasing", space="lch"),
            Color("lch(32.393 61.244 252.89)")
        )
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.25, hue="specified", space="lch"),
            Color("lch(32.393 61.244 432.89)")
        )

    def test_hue_shorter_cases(self):
        """Cover shorter hue cases."""

        # c2 - c1 > 180
        c1 = Color('lch(75% 50 40)')
        c2 = Color('lch(30% 30 350)')
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.50, hue="shorter", space="lch"),
            Color("lch(75% 50 15)")
        )

        # c2 - c1 < -180
        c1 = Color('lch(30% 30 350)')
        c2 = Color('lch(75% 50 40)')
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.50, hue="shorter", space="lch"),
            Color("lch(30% 30 375)")
        )

    def test_hue_longer_cases(self):
        """Cover longer hue cases."""

        # 0 < (c2 - c1) < 180
        c1 = Color('lch(75% 50 40)')
        c2 = Color('lch(30% 30 60)')
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.50, hue="longer", space="lch"),
            Color("lch(75% 50 -130)")
        )

        # -180 < (c2 - c1) < 0
        c1 = Color('lch(30% 30 60)')
        c2 = Color('lch(75% 50 40)')
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.50, hue="longer", space="lch"),
            Color("lch(30% 30 230)")
        )

    def test_hue_increasing_cases(self):
        """Cover increasing hue cases."""

        # c2 < c1
        c1 = Color('lch(75% 50 60)')
        c2 = Color('lch(30% 30 40)')
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.50, hue="increasing", space="lch"),
            Color("lch(75% 50 230)")
        )

    def test_hue_decreasing_cases(self):
        """Cover decreasing hue cases."""

        # c1 < c2
        c1 = Color('lch(75% 50 40)')
        c2 = Color('lch(30% 30 60)')
        self.assertColorEqual(
            c1.mix(c2.mask("hue", invert=True), 0.50, hue="decreasing", space="lch"),
            Color("lch(75% 50 -130)")
        )

    def test_mix_hue_adjust_bad(self):
        """Test hue adjusting."""

        c1 = Color('rebeccapurple')
        c2 = Color('lch(85% 100 805)')
        with self.assertRaises(ValueError):
            c1.mix(c2.mask("hue", invert=True), 0.25, hue="bad", space="lch")

    def test_mix_hue_nan(self):
        """Test mix hue with `NaN`."""

        self.assertColorEqual(
            Color('hsl', [NaN, 0, 0.25]).mix(Color('hsl', [NaN, 0, 0.9]), 0.50, space="hsl"),
            Color("hsl(0, 0%, 57.5%)")
        )

        self.assertColorEqual(
            Color('hsl', [NaN, 0, 0.25]).mix(Color('hsl', [120, 0.5, 0.9]), 0.50, space="hsl"),
            Color("hsl(120, 25%, 57.5%)")
        )

        self.assertColorEqual(
            Color('hsl', [120, 0.5, 0.25]).mix(Color('hsl', [NaN, 0, 0.9]), 0.50, space="hsl"),
            Color("hsl(120, 25%, 57.5%)")
        )

    def test_mix_progress(self):
        """Test custom progress."""

        results = [
            Color('color(--lab 54.291 80.805 69.891 / 1)'),
            Color('color(--lab 51.98 79.635 52.888 / 1)'),
            Color('color(--lab 46.494 76.857 12.521 / 1)'),
            Color('color(--lab 38.917 73.021 -43.239 / 1)'),
            Color('color(--lab 29.568 68.287 -112.03 / 1)')
        ]
        self.assertColorEqual(
            Color('red').mix('blue', 1, out_space="lab", space="lab", progress=ease_in),
            results[4]
        )
        self.assertColorEqual(
            Color('red').mix('blue', 0.75, out_space="lab", space="lab", progress=ease_in),
            results[3]
        )
        self.assertColorEqual(
            Color('red').mix('blue', 0.5, out_space="lab", space="lab", progress=ease_in),
            results[2]
        )
        self.assertColorEqual(
            Color('red').mix('blue', 0.25, out_space="lab", space="lab", progress=ease_in),
            results[1]
        )
        self.assertColorEqual(
            Color('red').mix('blue', 0, out_space="lab", space="lab", progress=ease_in),
            results[0]
        )

    def test_interpolate(self):
        """Test interpolation."""

        self.assertColorEqual(Color.interpolate(['red', 'blue'], space="srgb")(1), Color("srgb", [0, 0, 1]))
        self.assertColorEqual(Color.interpolate(['red', 'blue'], space="srgb")(0.75), Color("srgb", [0.25, 0, 0.75]))
        self.assertColorEqual(Color.interpolate(['red', 'blue'], space="srgb")(0.5), Color("srgb", [0.5, 0, 0.5]))
        self.assertColorEqual(Color.interpolate(['red', 'blue'], space="srgb")(0.25), Color("srgb", [0.75, 0, 0.25]))
        self.assertColorEqual(Color.interpolate(['red', 'blue'], space="srgb")(0), Color("srgb", [1, 0, 0]))

    def test_interpolate_bspline(self):
        """Test interpolation B-spline."""

        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], space="srgb", method="bspline")(1), Color("srgb", [0, 0, 1])
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], space="srgb", method="bspline")(0.75), Color("srgb", [0.25, 0, 0.75])
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], space="srgb", method="bspline")(0.5), Color("srgb", [0.5, 0, 0.5])
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], space="srgb", method="bspline")(0.25), Color("srgb", [0.75, 0, 0.25])
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], space="srgb", method="bspline")(0), Color("srgb", [1, 0, 0])
        )

    def test_interpolate_channel(self):
        """Test interpolating a specific channel differently."""

        self.assertColorEqual(
            Color.interpolate(['red', Color('blue').set('alpha', 0)], progress={'alpha': lambda t: t ** 3})(0.5),
            Color('oklab(0.35883 0.12849 0.07191 / 0.875)')
        )

    def test_interpolate_channel_bspline(self):
        """Test interpolating a specific channel differently."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0)], progress={'alpha': lambda t: t ** 3}, method='bspline'
            )(0.5),
            Color('oklab(0.35883 0.12849 0.07191 / 0.875)')
        )

    def test_interpolate_easing_inline(self):
        """Test interpolating a specific channel differently."""

        self.assertColorEqual(
            Color.interpolate(['red', lambda t: t ** 3, 'blue'])(0.5),
            Color('oklab(0.60596 0.1927 0.07117)')
        )

    def test_interpolate_color_hint(self):
        """Test interpolating with color hints."""

        self.assertColorEqual(
            Color.interpolate(['red', hint(0.75), 'blue'])(0.5),
            Color('oklab(0.59484 0.17643 0.04352)')
        )

    def test_interpolate_channel_all(self):
        """Test interpolating a specific channel differently, but setting the others via all."""

        self.assertColorEqual(
            Color.interpolate(
                ['red', Color('blue').set('alpha', 0)],
                progress={
                    'alpha': lambda t: t ** 3,
                    'all': lambda t: 0,
                },
                space='srgb'
            )(0.5),
            Color('rgb(291.43 0 0 / 0.875)')
        )

    def test_interpolate_channel_aliases(self):
        """Test interpolating a specific channel using a color's channel alias."""

        self.assertColorEqual(
            Color.interpolate(
                ['orange', Color('purple')],
                progress={
                    'red': lambda t: t ** 3
                },
                space='srgb'
            )(0.5),
            Color('rgb(191.5 82.5 64)')
        )

    def test_interpolate_input_piecewise(self):
        """Test interpolation with piecewise."""

        self.assertColorEqual(
            Color.interpolate(['red', stop('blue', 0.5)], space="srgb")(0.5), Color("srgb", [0, 0, 1])
        )

    def test_interpolate_input_bspline(self):
        """Test interpolation with piecewise."""

        self.assertColorEqual(
            Color.interpolate(['red', stop('blue', 0.5)], space="srgb", method='bspline')(0.5),
            Color("srgb", [0, 0, 1])
        )

    def test_interpolate_stop(self):
        """Test interpolation with piecewise."""

        self.assertColorEqual(
            Color.interpolate([stop('red', 0.6), 'blue'], space="srgb")(0.5), Color('red')
        )
        self.assertColorEqual(
            Color.interpolate([stop('red', 0.6), 'blue'], space="srgb")(0.7), Color('rgb(191.25 0 63.75)')
        )

    def test_interpolate_stop_bspline(self):
        """Test interpolation with piecewise."""

        self.assertColorEqual(
            Color.interpolate([stop('red', 0.6), 'blue'], space="srgb", method='bspline')(0.5), Color('red')
        )

    def test_interpolate_space(self):
        """Test color mix in different space."""

        self.assertColorEqual(Color.interpolate(['red', 'blue'], space='lab')(1), Color("lab(29.568 68.287 -112.03)"))
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], space='lab')(0.75), Color("lab(35.749 71.417 -66.55)")
        )
        self.assertColorEqual(Color.interpolate(['red', 'blue'], space='lab')(0.5), Color("lab(41.929 74.546 -21.069)"))
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], space='lab')(0.25), Color("lab(48.11 77.676 24.411)")
        )
        self.assertColorEqual(Color.interpolate(['red', 'blue'], space='lab')(0), Color("lab(54.291 80.805 69.891)"))

    def test_interpolate_empty_list(self):
        """Test interpolate with empty list."""

        with self.assertRaises(IndexError):
            Color('green').interpolate([])(0.5)

    def test_interpolate_piecewise(self):
        """Test multiple inputs for interpolation."""

        func = Color.interpolate(['white', 'red', 'black'])
        self.assertColorEqual(func(0), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(0.25), Color('color(--oklab 0.81398 0.11243 0.06292 / 1)'))
        self.assertColorEqual(func(0.5), Color('color(--oklab 0.62796 0.22486 0.12585 / 1)'))
        self.assertColorEqual(func(0.75), Color('color(--oklab 0.31398 0.11243 0.06292 / 1)'))
        self.assertColorEqual(func(1), Color('color(--oklab 0 0 0 / 1)'))
        self.assertColorEqual(func(-0.1), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(1.1), Color('color(--oklab 0 0 0 / 1)'))

    def test_interpolate_multi_bspline(self):
        """Test multiple inputs for B-spline interpolation."""

        func = Color.interpolate(['white', 'red', 'black'], method='bspline')
        self.assertColorEqual(func(0), Color('oklab(1 0 0)'))
        self.assertColorEqual(func(0.25), Color('color(--oklab 0.80865 0.10306 0.05768 / 1)'))
        self.assertColorEqual(func(0.5), Color('color(--oklab 0.5853 0.14991 0.0839 / 1)'))
        self.assertColorEqual(func(0.75), Color('color(--oklab 0.30865 0.10306 0.05768 / 1)'))
        self.assertColorEqual(func(1), Color('color(--oklab 0 0 0 / 1)'))
        self.assertColorEqual(func(-0.1), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(1.1), Color('color(--oklab 0 0 0 / 1)'))

    def test_interpolate_multi_natural(self):
        """Test multiple inputs for Natural B-spline interpolation."""

        func = Color.interpolate(['white', 'red', 'black'], method='natural')
        self.assertColorEqual(func(0), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(0.25), Color('color(--oklab 0.83797 0.15459 0.08652 / 1)'))
        self.assertColorEqual(func(0.5), Color('color(--oklab 0.62796 0.22486 0.12585 / 1)'))
        self.assertColorEqual(func(0.75), Color('color(--oklab 0.33797 0.15459 0.08652 / 1)'))
        self.assertColorEqual(func(1), Color('color(--oklab 0 0 0 / 1)'))
        self.assertColorEqual(func(-0.1), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(1.1), Color('color(--oklab 0 0 0 / 1)'))

    def test_interpolate_multi_natural_more(self):
        """Test more inputs for Natural B-spline interpolation."""

        func = Color.interpolate(['white', 'red', 'black', 'purple'], method='natural')
        self.assertColorEqual(func(0), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(0.25), Color('color(--oklab 0.7663 0.21652 0.11694 / 1)'))
        self.assertColorEqual(func(0.5), Color('color(--oklab 0.25451 0.11694 0.07997 / 1)'))
        self.assertColorEqual(func(0.75), Color('color(--oklab 0.00785 -0.00275 -0.03301 / 1)'))
        self.assertColorEqual(func(1), Color('color(--oklab 0.42091 0.1647 -0.10147 / 1)'))
        self.assertColorEqual(func(-0.1), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(1.1), Color('color(--oklab 0.42091 0.1647 -0.10147 / 1)'))

    def test_interpolate_multi_natural_even_more(self):
        """Test even more inputs for Natural B-spline interpolation."""

        func = Color.interpolate(['white', 'red', 'black', 'purple', 'green'], method='natural')
        self.assertColorEqual(func(0), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(0.25), Color('color(--oklab 0.62796 0.22486 0.12585 / 1)'))
        self.assertColorEqual(func(0.5), Color('color(--oklab 0 0 0 / 1)'))
        self.assertColorEqual(func(0.75), Color('color(--oklab 0.42091 0.1647 -0.10147 / 1)'))
        self.assertColorEqual(func(1), Color('color(--oklab 0.51975 -0.1403 0.10768 / 1)'))
        self.assertColorEqual(func(-0.1), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(1.1), Color('color(--oklab 0.51975 -0.1403 0.10768 / 1)'))

    def test_interpolate_multi_catmull_rom(self):
        """Test multiple inputs for Catmull-Rom spline interpolation."""

        func = Color.interpolate(['white', 'red', 'black'], method='catrom')
        self.assertColorEqual(func(0), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(0.25), Color('color(--oklab 0.82997 0.14054 0.07865 / 1)'))
        self.assertColorEqual(func(0.5), Color('color(--oklab 0.62796 0.22486 0.12585 / 1)'))
        self.assertColorEqual(func(0.75), Color('color(--oklab 0.32997 0.14054 0.07865 / 1)'))
        self.assertColorEqual(func(1), Color('color(--oklab 0 0 0 / 1)'))
        self.assertColorEqual(func(-0.1), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(1.1), Color('color(--oklab 0 0 0 / 1)'))

    def test_interpolate_multi_monotone(self):
        """Test multiple inputs for monotone interpolation."""

        func = Color.interpolate(['white', 'red', 'black'], method='monotone')
        self.assertColorEqual(func(0), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(0.25), Color('color(--oklab 0.82997 0.14054 0.07865 / 1)'))
        self.assertColorEqual(func(0.5), Color('color(--oklab 0.62796 0.22486 0.12585 / 1)'))
        self.assertColorEqual(func(0.75), Color('color(--oklab 0.32997 0.14054 0.07865 / 1)'))
        self.assertColorEqual(func(1), Color('color(--oklab 0 0 0 / 1)'))
        self.assertColorEqual(func(-0.1), Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(func(1.1), Color('color(--oklab 0 0 0 / 1)'))

    def test_interpolate_out_space(self):
        """Test interpolation."""

        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab")(1),
            Color("lab(29.568% 68.287 -112.03)")
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab")(0.75),
            Color("lab(35.749% 71.417 -66.55)")
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab")(0.5),
            Color("lab(41.929% 74.546 -21.069)")
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab")(0.25),
            Color("lab(48.11% 77.676 24.411)")
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab")(0),
            Color("lab(54.291% 80.805 69.891)")
        )

    def test_interpolate_alpha(self):
        """Test mixing alpha."""

        self.assertColorEqual(
            Color.interpolate(['color(srgb 1 0 0 / 0.75)', 'color(srgb 0 0 1 / 0.25)'], space="srgb")(0.5),
            Color('rgb(191.25 0 63.75 / 0.5)')
        )

    def test_interpolate_premultiplied_alpha(self):
        """Test premultiplied alpha."""

        self.assertColorEqual(
            Color.interpolate(
                ['color(srgb 1 0 0 / 0.75)', 'color(srgb 0 0 1 / 0.25)'], space="srgb", premultiplied=True
            )(0.5),
            Color('rgb(191.25 0 63.75 / 0.5)')
        )

    def test_interpolate_premultiplied_alpha_none(self):
        """Test premultiplied alpha when alphas are none."""

        self.assertColorEqual(
            Color.interpolate(
                ['color(srgb 0 0 0 / none)', 'color(srgb 0 1 0 / none)'], space="srgb", premultiplied=True
            )(0.5),
            Color('rgb(0 127.5 0 / 0)')
        )

    def test_interpolate_premultiplied_no_alpha(self):
        """Test premultiplied alpha."""

        self.assertColorEqual(
            Color.interpolate(['color(srgb 1 0 0)', 'color(srgb 0 0 1)'], space="srgb", premultiplied=True)(0.5),
            Color.interpolate(['color(srgb 1 0 0)', 'color(srgb 0 0 1)'], space="srgb")(0.5)
        )

    def test_interpolate_nan(self):
        """Test mixing with NaN."""

        c1 = Color("srgb", [NaN, 1, 1])
        c2 = Color("srgb", [0.75, 0, 0])
        self.assertColorEqual(Color.interpolate([c1, c2], space="srgb")(0.5), Color("srgb", [0.75, 0.5, 0.5]))
        c1 = Color("srgb", [0.25, 1, 1])
        c2 = Color("srgb", [NaN, 0, 0])
        self.assertColorEqual(Color.interpolate([c1, c2], space="srgb")(0.5), Color("srgb", [0.25, 0.5, 0.5]))
        c1 = Color("srgb", [NaN, 1, 1])
        c2 = Color("srgb", [NaN, 0, 0])
        self.assertColorEqual(Color.interpolate([c1, c2], space="srgb")(0.5), Color("srgb", [0, 0.5, 0.5]))

    def test_interpolate_adjust(self):
        """Test mix adjust method."""

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(
            Color.interpolate([c1, c2.mask("red")], space="srgb")(0.5),
            Color("srgb", [0.25, 0.5, 0.5])
        )

    def test_interpolate_hue_adjust(self):
        """Test hue adjusting."""

        c1 = Color('rebeccapurple')
        c2 = Color('lch(85% 100 805)')
        self.assertColorEqual(
            Color.interpolate([c1, c2.mask("hue", invert=True)], hue="shorter", space="lch")(0.25),
            Color("lch(32.393 61.244 342.89)")
        )
        self.assertColorEqual(
            Color.interpolate([c1, c2.mask("hue", invert=True)], hue="longer", space="lch")(0.25),
            Color("lch(32.393 61.244 252.89)")
        )
        self.assertColorEqual(
            Color.interpolate([c1, c2.mask("hue", invert=True)], hue="increasing", space="lch")(0.25),
            Color("lch(32.393 61.244 342.89)")
        )
        self.assertColorEqual(
            Color.interpolate([c1, c2.mask("hue", invert=True)], hue="decreasing", space="lch")(0.25),
            Color("lch(32.393 61.244 252.89)")
        )
        self.assertColorEqual(
            Color.interpolate([c1, c2.mask("hue", invert=True)], hue="specified", space="lch")(0.25),
            Color("lch(32.393 61.244 432.89)")
        )

    def test_interpolate_progress(self):
        """Test custom progress."""

        results = [
            Color('color(--lab 54.291 80.805 69.891 / 1)'),
            Color('color(--lab 51.98 79.635 52.888 / 1)'),
            Color('color(--lab 46.494 76.857 12.521 / 1)'),
            Color('color(--lab 38.917 73.021 -43.239 / 1)'),
            Color('color(--lab 29.568 68.287 -112.03 / 1)')
        ]
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab", progress=ease_in)(1),
            results[4]
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab", progress=ease_in)(0.75),
            results[3]
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab", progress=ease_in)(0.5),
            results[2]
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab", progress=ease_in)(0.25),
            results[1]
        )
        self.assertColorEqual(
            Color.interpolate(['red', 'blue'], out_space="lab", space="lab", progress=ease_in)(0),
            results[0]
        )

    def test_steps(self):
        """Test steps."""

        colors = Color.steps(['red', 'blue'], space="srgb", steps=5)
        self.assertColorEqual(colors[4], Color("srgb", [0, 0, 1]))
        self.assertColorEqual(colors[3], Color("srgb", [0.25, 0, 0.75]))
        self.assertColorEqual(colors[2], Color("srgb", [0.5, 0, 0.5]))
        self.assertColorEqual(colors[1], Color("srgb", [0.75, 0, 0.25]))
        self.assertColorEqual(colors[0], Color("srgb", [1, 0, 0]))

    def test_steps_input_piecewise(self):
        """Test steps with piecewise."""

        self.assertColorEqual(
            Color.steps(['red', stop('blue', 0.5)], space="srgb", steps=5)[2], Color("srgb", [0, 0, 1])
        )

    def test_steps_input_bspline(self):
        """Test steps with B-spline."""

        self.assertColorEqual(
            Color.steps(['red', stop('blue', 0.5)], space="srgb", steps=5, method='bspline')[2],
            Color("srgb", [0, 0, 1])
        )

    def test_steps_multi(self):
        """Test steps with multiple color ranges."""

        colors = Color('white').steps(['white', 'red', 'black'], steps=3)
        self.assertColorEqual(colors[0], Color('color(--oklab 1 0 0 / 1)'))
        self.assertColorEqual(colors[1], Color('oklab(0.62796 0.22486 0.12585)'))
        self.assertColorEqual(colors[2], Color('color(--oklab 0 0 0 / 1)'))

    def test_steps_multi_max_delta_e(self):
        """Test steps with multiple color ranges and max_delta_e."""

        colors = Color.steps(['red', 'green', 'blue'], space="srgb", max_delta_e=10)
        for index, color in enumerate(colors, 0):
            if not index:
                continue
            self.assertTrue(color.delta_e(colors[index - 1]) <= 10)

        colors = Color.steps(['red', 'green', 'blue'], space="srgb", max_delta_e=3)
        for index, color in enumerate(colors, 0):
            if not index:
                continue
            self.assertTrue(color.delta_e(colors[index - 1]) <= 3)

    def test_steps_custom_delta_e(self):
        """Test a custom delta E input."""

        colors = Color.steps(['red', 'green', 'blue'], space="srgb", max_delta_e=10, delta_e='2000')
        for index, color in enumerate(colors, 0):
            if not index:
                continue
            self.assertTrue(color.delta_e(colors[index - 1], method='2000') <= 10)

        colors = Color.steps(['red', 'green', 'blue'], space="srgb", max_delta_e=3, delta_e='2000')
        for index, color in enumerate(colors, 0):
            if not index:
                continue
            self.assertTrue(color.delta_e(colors[index - 1], method='2000') <= 3)

    def test_steps_custom_delta_compare(self):
        """Test a custom delta E input."""

        colors1 = len(Color.steps(['orange', 'red'], space="srgb", max_delta_e=0.2, delta_e='76'))
        colors2 = len(Color.steps(['orange', 'red'], space="srgb", max_delta_e=0.2, delta_e='ok'))

        self.assertNotEqual(colors1, colors2)

    def test_steps_empty_list(self):
        """Test steps with empty list."""

        with self.assertRaises(IndexError):
            Color.steps([], steps=3)

    def test_steps_space(self):
        """Test steps different space."""

        colors = Color.steps(['red', 'blue'], space="lab", steps=5)
        self.assertColorEqual(colors[4], Color("lab(29.568 68.287 -112.03)"))
        self.assertColorEqual(colors[3], Color("lab(35.749 71.417 -66.55)"))
        self.assertColorEqual(colors[2], Color("lab(41.929 74.546 -21.069)"))
        self.assertColorEqual(colors[1], Color("lab(48.11 77.676 24.411)"))
        self.assertColorEqual(colors[0], Color("lab(54.291 80.805 69.891)"))

    def test_steps_out_space(self):
        """Test steps with output in different space."""

        colors = Color.steps(['red', 'blue'], space="srgb", steps=5, out_space="lab")
        self.assertColorEqual(
            colors[4],
            Color("lab(29.568% 68.287 -112.03)")
        )
        self.assertColorEqual(
            colors[3],
            Color("lab(24.638% 57.331 -83.546)")
        )
        self.assertColorEqual(
            colors[2],
            Color("lab(29.563% 55.954 -36.19)")
        )
        self.assertColorEqual(
            colors[1],
            Color("lab(41.111% 66.202 23.413)")
        )
        self.assertColorEqual(
            colors[0],
            Color("lab(54.291% 80.805 69.891)")
        )

    def test_steps_alpha(self):
        """Test mixing alpha."""

        self.assertColorEqual(
            Color.steps(['color(srgb 1 0 0 / 0.75)', 'color(srgb 0 0 1 / 0.25)'], space="srgb", steps=1)[0],
            Color('rgb(191.25 0 63.75 / 0.5)')
        )

    def test_steps_premultiplied_alpha(self):
        """Test premultiplied alpha."""

        self.assertColorEqual(
            Color.steps(
                ['color(srgb 1 0 0 / 0.75)', 'color(srgb 0 0 1 / 0.25)'], space="srgb", steps=1, premultiplied=True
            )[0],
            Color('rgb(191.25 0 63.75 / 0.5)')
        )

    def test_steps_premultiplied_no_alpha(self):
        """Test premultiplied alpha."""

        self.assertColorEqual(
            Color.steps(['color(srgb 1 0 0)', 'color(srgb 0 0 1)'], space="srgb", steps=1, premultiplied=True)[0],
            Color.steps(['color(srgb 1 0 0)', 'color(srgb 0 0 1)'], space="srgb", steps=1)[0]
        )

    def test_steps_nan(self):
        """Test steps with NaN."""

        c1 = Color("srgb", [NaN, 1, 1])
        c2 = Color("srgb", [0.75, 0, 0])
        self.assertColorEqual(Color.steps([c1, c2], space="srgb", steps=1)[0], Color("srgb", [0.75, 0.5, 0.5]))
        c1 = Color("srgb", [0.25, 1, 1])
        c2 = Color("srgb", [NaN, 0, 0])
        self.assertColorEqual(Color.steps([c1, c2], space="srgb", steps=1)[0], Color("srgb", [0.25, 0.5, 0.5]))
        c1 = Color("srgb", [NaN, 1, 1])
        c2 = Color("srgb", [NaN, 0, 0])
        self.assertColorEqual(Color.steps([c1, c2], space="srgb", steps=1)[0], Color("srgb", [0, 0.5, 0.5]))

    def test_steps_adjust(self):
        """Test steps with adjust method."""

        c1 = Color("color(srgb 0.25 1 1)")
        c2 = Color("color(srgb 0.75 0 0)")
        self.assertColorEqual(
            Color.steps([c1, c2.mask("red")], space="srgb", steps=1)[0],
            Color("srgb", [0.25, 0.5, 0.5])
        )

    def test_steps_hue_adjust(self):
        """Test steps with hue adjusting."""

        self.assertColorEqual(
            Color.steps(
                ['rebeccapurple', Color('lch(85% 100 805)').mask("hue", invert=True)],
                space="lch",
                steps=5,
                hue="shorter"
            )[1],
            Color("lch(32.393 61.244 342.89)")
        )
        self.assertColorEqual(
            Color.steps(
                ['rebeccapurple', Color('lch(85% 100 805)').mask("hue", invert=True)],
                space="lch",
                steps=5,
                hue="longer"
            )[1],
            Color("lch(32.393 61.244 252.89)")
        )
        self.assertColorEqual(
            Color.steps(
                ['rebeccapurple', Color('lch(85% 100 805)').mask("hue", invert=True)],
                space="lch",
                steps=5,
                hue="increasing"
            )[1],
            Color("lch(32.393 61.244 342.89)")
        )
        self.assertColorEqual(
            Color.steps(
                ['rebeccapurple', Color('lch(85% 100 805)').mask("hue", invert=True)],
                space="lch",
                steps=5,
                hue="decreasing"
            )[1],
            Color("lch(32.393 61.244 252.89)")
        )
        self.assertColorEqual(
            Color.steps(
                ['rebeccapurple', Color('lch(85% 100 805)').mask("hue", invert=True)],
                space="lch",
                steps=5,
                hue="specified"
            )[1],
            Color("lch(32.393 61.244 432.89)")
        )

    def test_steps_progress(self):
        """Test custom progress."""

        results = [
            Color('color(--lab 54.291 80.805 69.891 / 1)'),
            Color('color(--lab 51.98 79.635 52.888 / 1)'),
            Color('color(--lab 46.494 76.857 12.521 / 1)'),
            Color('color(--lab 38.917 73.021 -43.239 / 1)'),
            Color('color(--lab 29.568 68.287 -112.03 / 1)')
        ]
        colors = Color.steps(['red', 'blue'], steps=5, out_space="lab", space="lab", progress=ease_in)
        self.assertColorEqual(
            colors[4],
            results[4]
        )
        self.assertColorEqual(
            colors[3],
            results[3]
        )
        self.assertColorEqual(
            colors[2],
            results[2]
        )
        self.assertColorEqual(
            colors[1],
            results[1]
        )
        self.assertColorEqual(
            colors[0],
            results[0]
        )

    def test_steps_max_delta_e(self):
        """Test steps with a max delta e."""

        colors = Color.steps(['red', 'blue'], space="srgb", max_delta_e=10)
        for index, color in enumerate(colors, 0):
            if not index:
                continue
            self.assertTrue(color.delta_e(colors[index - 1]) <= 10)

        colors = Color.steps(['red', 'blue'], space="srgb", max_delta_e=3)
        for index, color in enumerate(colors, 0):
            if not index:
                continue
            self.assertTrue(color.delta_e(colors[index - 1]) <= 3)

    def test_max_delta_min_step_less_than_two(self):
        """Test that when a minimum step less than 2 is given that `max_delta_e` won't break."""

        colors = Color.steps(['lightblue', 'blue'], space="srgb", steps=1, max_delta_e=10)
        self.assertTrue(len(colors) > 2)

    def test_steps_max_delta_e_steps(self):
        """Test steps with a max delta e."""

        colors = Color.steps(['red', 'blue'], space="srgb", max_delta_e=10)
        self.assertTrue(len(colors) > 5)
        colors = Color.steps(['red', 'blue'], space="srgb", max_delta_e=10, max_steps=5)
        self.assertTrue(len(colors) == 5)

    def test_too_few_colors_linear(self):
        """Test too few colors during linear interpolation."""

        with self.assertRaises(ValueError):
            Color.interpolate(['green', lambda t: t * 3])

    def test_continuous_undefined_middle(self):
        """Test continuous with undefined middle."""

        colors = [
            Color('oklab', [0, 0, 0]),
            Color('oklab', [NaN, -0.03246, -0.31153]),
            Color('oklab', [1, 0, 0])
        ]
        i = Color.interpolate(colors, space='oklab', method='continuous')
        self.assertColorEqual(
            i(0),
            Color("oklab(0 0 0)")
        )
        self.assertColorEqual(
            i(0.25),
            Color("oklab(0.25 -0.01623 -0.15576)")
        )
        self.assertColorEqual(
            i(0.5),
            Color("oklab(0.5 -0.03246 -0.31153)")
        )
        self.assertColorEqual(
            i(0.75),
            Color("oklab(0.75 -0.01623 -0.15576)")
        )
        self.assertColorEqual(
            i(1),
            Color("oklab(1 0 0)")
        )

    def test_continuous_undefined_left(self):
        """Test continuous with undefined left."""

        colors = [
            Color('oklab', [NaN, 0.22486, 0.12585]),
            Color('oklab', [NaN, -0.1403, 0.10768]),
            Color('oklab', [0.45201, -0.03246, -0.31153])
        ]
        i = Color.interpolate(colors, space='oklab', method='continuous')
        self.assertColorEqual(
            i(0),
            Color("oklab(0.45201 0.22486 0.12585)")
        )
        self.assertColorEqual(
            i(0.25),
            Color("oklab(0.45201 0.04228 0.11677)")
        )
        self.assertColorEqual(
            i(0.5),
            Color("oklab(0.45201 -0.1403 0.10768)")
        )
        self.assertColorEqual(
            i(0.75),
            Color("oklab(0.45201 -0.08638 -0.10192)")
        )
        self.assertColorEqual(
            i(1),
            Color("oklab(0.45201 -0.03246 -0.31153)")
        )

    def test_continuous_undefined_right(self):
        """Test continuous with undefined right."""

        colors = [
            Color('oklab', [0.45201, 0.22486, 0.12585]),
            Color('oklab', [NaN, -0.1403, 0.10768]),
            Color('oklab', [NaN, -0.03246, -0.31153])
        ]
        i = Color.interpolate(colors, space='oklab', method='continuous')
        self.assertColorEqual(
            i(0),
            Color("oklab(0.45201 0.22486 0.12585)")
        )
        self.assertColorEqual(
            i(0.25),
            Color("oklab(0.45201 0.04228 0.11677)")
        )
        self.assertColorEqual(
            i(0.5),
            Color("oklab(0.45201 -0.1403 0.10768)")
        )
        self.assertColorEqual(
            i(0.75),
            Color("oklab(0.45201 -0.08638 -0.10192)")
        )
        self.assertColorEqual(
            i(1),
            Color("oklab(0.45201 -0.03246 -0.31153)")
        )

    def test_too_few_colors_bspline(self):
        """Test too few colors during B-spline interpolation."""

        with self.assertRaises(ValueError):
            Color.interpolate(['green', lambda t: t * 3], method='bspline')

    def test_bad_method(self):
        """Test bad interpolation method."""

        with self.assertRaises(ValueError):
            Color.interpolate(['green', lambda t: t * 3], method='bad')

    def test_bad_easing(self):
        """Test bad color easing linear."""

        with self.assertRaises(ValueError):
            Color.interpolate([lambda t: t * 3, 'green'])

    def test_bad_color_input_bspline(self):
        """Test bad color easing B-spline."""

        with self.assertRaises(ValueError):
            Color.interpolate([lambda t: t * 3, 'green'], method='bspline')

    def test_bspline_all_none(self):
        """Test multiple B-spline inputs with the same channel all none."""

        self.assertColorEqual(
            Color.interpolate(
                ['Color(srgb 1 none 0.5)', 'Color(srgb 0.1 none 0.7)', 'Color(srgb 0.2 none 0.9)'],
                space='srgb',
                method='bspline'
            )(0.5),
            Color('rgb(68 none 178.5)')
        )

    def test_bspline_most_none(self):
        """Test multiple B-spline inputs with the same channel with most none."""

        self.assertColorEqual(
            Color.interpolate(
                ['Color(srgb 1 none 0.5)', 'Color(srgb 0.1 none 0.7)', 'Color(srgb 0.2 0.8 0.9)'],
                space='srgb',
                method='bspline'
            )(0.5),
            Color('rgb(68 204 178.5)')
        )

    def test_bspline_none_after(self):
        """Test multiple B-spline inputs with the same channel with most none but not none at the start."""

        self.assertColorEqual(
            Color.interpolate(
                ['Color(srgb 1 0.8 0.5)', 'Color(srgb 0.1 none 0.7)', 'Color(srgb 0.2 none 0.9)'],
                space='srgb',
                method='bspline'
            )(0.5),
            Color('rgb(68 204 178.5)')
        )

    def test_bspline_cylindrical(self):
        """Test B-spline with a cylindrical space."""

        self.assertColorEqual(
            Color.interpolate(
                ['hsl(250 50% 30%)', 'hsl(none 0% 10%)', 'hsl(120 75% 75%)'],
                space='hsl',
                method='bspline'
            )(0.75),
            Color('hsl(152.5 40.104% 44.271%)')
        )

    def test_bspline_undefined_alpha(self):
        """Test some cases related to undefined alpha and B-spline."""

        self.assertColorEqual(
            Color.interpolate(
                ['color(srgb 1 0 0 / none)', 'color(srgb 0 1 0 / 0.5)', 'color(srgb 0 0 1 / 1)'],
                space='srgb',
                method='bspline'
            )(0.75),
            Color('rgb(3.4932 38.425 174.66 / 0.76042)')
        )

        self.assertColorEqual(
            Color.interpolate(
                ['color(srgb 1 0 0 / 1)', 'color(srgb 0 1 0 / 0.5)', 'color(srgb 0 0 1 / none)'],
                space='srgb',
                method='bspline'
            )(0.75),
            Color('rgb(10.408 57.245 130.1 / 0.51042)')
        )

        self.assertColorEqual(
            Color.interpolate(
                ['color(srgb 1 0 0 / 1)', 'color(srgb 0 1 0 / none)', 'color(srgb 0 0 1 / 0.5)'],
                space='srgb',
                method='bspline'
            )(0.75),
            Color('rgb(8.5 140.25 106.25 / 0.625)')
        )

    def test_extrapolate_bspline(self):
        """
        Test that extrapolation of B-spline is linear after the first and last control point.

        Values have been vetted by plotting a complete curve that extrapolates past the endpoints.
        """

        i = Color.interpolate(
            ["oklab(0.7 0.15 0.1)", "oklab(0.7 -0.05 0.1)", "oklab(0.7 -0.09 0.02)", "oklab(0.7 -0.03 -0.12)"],
            method='bspline',
            extrapolate=True
        )
        self.assertColorEqual(i(-0.5), Color('oklab(0.7 0.25 0.1)'))
        self.assertColorEqual(i(1.5), Color('oklab(0.7 0 -0.19)'))

    def test_extrapolate_linear(self):
        """
        Test that extrapolation of linear makes sense.

        Values have been vetted by plotting a complete curve that extrapolates past the endpoints.
        """

        i = Color.interpolate(
            ["oklab(0.7 0.15 0.1)", "oklab(0.7 -0.05 0.1)", "oklab(0.7 -0.09 0.02)", "oklab(0.7 -0.03 -0.12)"],
            method='linear',
            extrapolate=True
        )
        self.assertColorEqual(i(-0.5), Color('color(--oklab 0.7 0.25 0.1 / 1)'))
        self.assertColorEqual(i(1.5), Color('color(--oklab 0.7 0 -0.19 / 1)'))

    def test_discrete(self):
        """Test discrete interpolation."""

        i = Color.discrete(
            ['blue', 'green', 'yellow', 'orange', 'red'],
            domain=[-32, 32, 60, 85, 95],
            space='srgb'
        )
        self.assertColorEqual(i(-20), Color('blue'))
        self.assertColorEqual(i(60), Color('yellow'))
        self.assertColorEqual(i(87), Color('orange'))
        self.assertColorEqual(i(100), Color('red'))

    def test_discrete_steps(self):
        """Test specifying discrete steps."""

        i = Color.discrete(
            ['blue', 'green', 'yellow', 'orange', 'red'],
            domain=[-32, 32, 60, 85, 95],
            steps=3,
            space='srgb'
        )
        self.assertColorEqual(i(-20), Color('blue'))
        self.assertColorEqual(i(60), Color('yellow'))
        self.assertColorEqual(i(87), Color('red'))
        self.assertColorEqual(i(100), Color('red'))

    def test_padding(self):
        """Test padding."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        i = Color.discrete(scale, space='srgb', padding=0.25)
        steps = 5
        colors = [i(r / (steps - 1)).to_string() for r in range(steps)]
        self.assertEqual(
            colors,
            ['rgb(253 212 158)',
             'rgb(253 187 132)',
             'rgb(252 141 89)',
             'rgb(239 101 72)',
             'rgb(215 48 31)']
        )

    def test_padding_output_space(self):
        """Test padding output space."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        i = Color.discrete(scale, space='srgb', padding=0.25, out_space='oklab')
        steps = 5
        colors = [i(r / (steps - 1)).to_string() for r in range(steps)]
        self.assertEqual(
            colors,
            ['oklab(0.89246 0.02285 0.08001)',
             'oklab(0.84091 0.05126 0.09064)',
             'oklab(0.7538 0.10664 0.10608)',
             'oklab(0.67477 0.14723 0.09751)',
             'oklab(0.57443 0.17728 0.10314)']
        )

    def test_padding_sequence_one(self):
        """Test padding sequence of one."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        i = Color.discrete(scale, space='srgb', padding=[0.25])
        steps = 5
        colors = [i(r / (steps - 1)).to_string() for r in range(steps)]
        self.assertEqual(
            colors,
            ['rgb(253 212 158)',
             'rgb(253 187 132)',
             'rgb(252 141 89)',
             'rgb(239 101 72)',
             'rgb(215 48 31)']
        )

    def test_padding_sequence_two(self):
        """Test padding sequence of two."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        i = Color.discrete(scale, space='srgb', padding=[0.25, 0.15])
        steps = 5
        colors = [i(r / (steps - 1)).to_string() for r in range(steps)]
        self.assertEqual(
            colors,
            ['rgb(253 212 158)',
             'rgb(252.8 177.8 123.4)',
             'rgb(246.8 125 82.2)',
             'rgb(224.6 69.2 47.4)',
             'rgb(186.2 9.6 6.2)']
        )

    def test_no_padding(self):
        """Test that 0 applies no padding."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        i = Color.discrete(scale, space='srgb', padding=0)
        steps = 5
        colors = [i(r / (steps - 1)).to_string() for r in range(steps)]
        self.assertEqual(
            colors,
            ['rgb(255 247 236)',
             'rgb(253 212 158)',
             'rgb(252 141 89)',
             'rgb(215 48 31)',
             'rgb(127 0 0)']
        )

    def test_padding_sequence_empty(self):
        """An empty list will be treated as no padding set at all."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        i = Color.discrete(scale, space='srgb', padding=[])
        steps = 5
        colors = [i(r / (steps - 1)).to_string() for r in range(steps)]
        self.assertEqual(
            colors,
            ['rgb(255 247 236)',
             'rgb(253 212 158)',
             'rgb(252 141 89)',
             'rgb(215 48 31)',
             'rgb(127 0 0)']
        )

    def test_padding_sequence_too_many(self):
        """Test padding with too many inputs."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        with self.assertRaises(ValueError):
            Color.discrete(scale, space='srgb', padding=[0.25, 0.15, 0.3])

    def test_padding_bad_output(self):
        """Test that a bad output for padding fails."""

        scale = ['#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59', '#ef6548', '#d7301f', '#b30000', '#7f0000']
        with self.assertRaises(ValueError):
            Color.discrete(scale, space='srgb', padding=[0.25, 0.15], out_space='bad')



"""
Easing functions.

https://drafts.csswg.org/css-easing/#easing-functions
https://probablymarcus.com/blocks/2015/02/26/using-bezier-curves-as-easing-functions.html
https://en.wikipedia.org/wiki/Horner%27s_method
https://en.wikipedia.org/wiki/Newton%27s_method

Since we know that control points P0 and P3 are always (0, 0) and (1, 1) respectively,
we can simplify the Bezier algorithm to:

```
x(t) = x0(1 - t)3 + 3x1(1 - t)2t + 3x2(1 - t)t2 + x3t3
x'(t) = 3(1 - t)2(x1 - x0) + 6(1 - t)t(x2 - x1) + 3t2(x3 - x2)
```

Further, they can be simplified to (Horner's method).

```
x(t) = (((1 - 3x2 + 3x1)t + 3x2 - 6x1)t + 3x1)t
x'(t) = 3(1 - 3x2 + 3x1)t2 + 2(3x2 - 6x1)t + 3x1
```

This allows us to calculate coefficients:

```
cx = 3.0 * p1x;
bx = 3.0 * (p2x - p1x) - cx;
ax = 1.0 - cx - bx;
```

And end up with:

```
x(t) = axt3 + bxt2 + cxt
x'(t) = 3axt2 + 2bxt + cx
```

This greatly simplifies things and makes it faster.
"""
from __future__ import annotations
import functools
from typing import Callable

EPSILON = 1e-6
# This value is honestly pretty arbitrary (was 10, now 8),
# probably would be good to test and find the optimal value.
MAX_ITER = 8


def _bezier(t: float, a: float, b: float, c: float) -> float:
    """
    Calculate the bezier point.

    We know that P0 and P3 are always (0, 0) and (1, 1) respectively.
    Knowing this we can simplify the equation by precalculating them in.
    """

    return a * t ** 3 + b * t ** 2 + c * t


def _derivative_x(t: float, a: float, b: float, c: float) -> float:
    """Derivative of curve."""

    return 3 * a * t ** 2 + 2 * b * t + c


def _solve_bezier_x(target: float, a: float, b: float, c: float) -> float:
    """
    Solve curve to find a `t` that satisfies our desired `x`.

    Try a few rounds of Newton's method which is generally faster. If we fail,
    resort to binary search.
    """

    # Try Newtons method to see if we can find a suitable value
    i = 0
    t = target
    while i < MAX_ITER:
        # See how close we are to the desired `x`
        x = _bezier(t, a, b, c) - target
        if abs(x) < EPSILON:
            return t

        # Get the derivative, but bail if it is too small,
        # we will just keep looping otherwise.
        d = _derivative_x(t, a, b, c)

        # Getting too close zero or is zero
        if abs(d) < EPSILON:  # pragma: no cover
            break

        # Calculate new time and try again
        t = t - x / d
        i += 1

    # Bisect until we arrive at a suitable value
    low, high = 0.0, 1.0
    t = target
    while abs(high - low) > EPSILON:
        x = _bezier(t, a, b, c)
        if (abs(x - target) < EPSILON):
            return t
        if x > target:
            high = t
        else:
            low = t
        t = (high + low) * 0.5

    # Just return whatever we got closest to
    # TODO: Is this needed? Can we actually ever hit this?
    return t  # pragma: no cover


def _extrapolate(t: float, p1: tuple[float, float], p2: tuple[float, float]) -> float:
    """
    Extrapolate.

    Try to use the closest tangent to the endpoint, but if we can't,
    we'll just end up returning the same `t`.
    """

    # See if we can use the tangent of a point closest to P3
    if t > 1:
        slope = 1.0

        # Use a straight line through P2 and P3
        if p2[0] < 1:
            slope = (p2[1] - 1.0) / (p2[0] - 1.0)

        # Use a straight line through P1 and P3
        elif p1[0] < 1:
            slope = (p1[1] - 1.0) / (p1[0] - 1.0)

        intercept = 1.0 - slope

    # See if we can use the tangent of a point closest to P0
    else:
        slope = 1.0

        # Use a straight line through P1 and P0
        if p1[0] > 0.0:
            slope = p1[1] / p1[0]

        # Use a straight line through P2 and P0
        elif p2[0] > 0:
            slope = p2[1] / p2[0]

        intercept = 0.0

    return slope * t + intercept


def _calc_bezier(
    target: float,
    a: tuple[float, float],
    b: tuple[float, float],
    c: tuple[float, float],
    p1: tuple[float, float],
    p2: tuple[float, float]
) -> float:
    """
    Calculate the y value of the bezier curve with the given `x`.

    The value given is actually the target `x`. We need to find a
    `t` that satisfies the `x` so that we can find the `y`.
    """

    # We'll likely not get a nice round 0 or 1 using the methods below,
    # but we know 0 and 1 should yield 0 and 1, so shortcut out and
    # same some calculations.
    if target in (0, 1):
        return target

    # Extrapolate per the spec
    if target > 1 or target < 0:
        return _extrapolate(target, p1, p2)

    # Solve for `t` in relation to `x`
    t = _solve_bezier_x(target, a[0], b[0], c[0])

    # Use the found `t` to locate the `y`
    y = _bezier(t, a[1], b[1], c[1])
    return y


def cubic_bezier(x1: float, y1: float, x2: float, y2: float) -> Callable[..., float]:
    """Return a cubic bezier easing function."""

    if x1 < 0 or x1 > 1 or x2 < 0 or x2 > 1:
        raise ValueError("Cubic Bezier requires 'x' values to be between 0 - 1")

    # Precalculate the coefficients for Horner's method
    cx = 3.0 * x1
    bx = 3.0 * (x2 - x1) - cx
    ax = 1.0 - cx - bx

    cy = 3.0 * y1
    by = 3.0 * (y2 - y1) - cy
    ay = 1.0 - cy - by

    return functools.partial(_calc_bezier, a=(ax, ay), b=(bx, by), c=(cx, cy), p1=(x1, y1), p2=(x2, y2))


def linear(t: float) -> float:
    """Linear."""

    return t


# CSS Easings Level 2
ease = cubic_bezier(0.25, 0.1, 0.25, 1.0)
ease_in = cubic_bezier(0.42, 0.0, 1.0, 1.0)
ease_out = cubic_bezier(0.0, 0.0, 0.58, 1.0)
ease_in_out = cubic_bezier(0.42, 0.0, 0.58, 1.0)



"""
HPLuv color space.

Adapted to Python and ColorAide by Isaac Muse (2021)

--- HSLuv Conversion Algorithm ---
Copyright (c) 2012-2021 Alexei Boronine
Copyright (c) 2016 Florian Dormont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations
from ..spaces import Space, HSLish
from ..cat import WHITES
from ..channels import Channel, FLG_ANGLE
from .lab import EPSILON, KAPPA
from .srgb_linear import XYZ_TO_RGB
import math
from .. import util
from ..types import Vector


def distance_line_from_origin(line: tuple[float, float]) -> float:
    """Distance line from origin."""

    return abs(line[1]) / math.sqrt(line[0] ** 2 + 1)


def get_bounds(l: float) -> list[tuple[float, float]]:
    """Get bounds."""

    result = []
    sub1 = ((l + 16) ** 3) / 1560896
    sub2 = sub1 if sub1 > EPSILON else l / KAPPA

    g = 0
    while g < 3:
        c = g
        g += 1
        m1, m2, m3 = XYZ_TO_RGB[c]
        g1 = 0
        while g1 < 2:
            t = g1
            g1 += 1
            top1 = (284517 * m1 - 94839 * m3) * sub2
            top2 = (838422 * m3 + 769860 * m2 + 731718 * m1) * l * sub2 - (769860 * t) * l
            bottom = (632260 * m3 - 126452 * m2) * sub2 + 126452 * t
            result.append((top1 / bottom, top2 / bottom))  # (slope, intercept)
    return result


def max_safe_chroma_for_l(l: float) -> float:
    """Get safe max chroma for lightness."""

    return min(distance_line_from_origin(bound) for bound in get_bounds(l))


def hpluv_to_lch(hpluv: Vector) -> Vector:
    """Convert HPLuv to LCh."""

    h, s, l = hpluv
    c = 0.0
    if l > 100 - 1e-7:
        l = 100
    elif l < 1e-08:
        l = 0.0
    else:
        _hx_max = max_safe_chroma_for_l(l)
        c = _hx_max / 100 * s
    return [l, c, util.constrain_hue(h)]


def lch_to_hpluv(lch: Vector) -> Vector:
    """Convert LCh to HPLuv."""

    l, c, h = lch
    s = 0.0
    if l > 100 - 1e-7:
        l = 100
    elif l < 1e-08:
        l = 0.0
    else:
        _hx_max = max_safe_chroma_for_l(l)
        s = c / _hx_max * 100
    return [util.constrain_hue(h), s, l]


class HPLuv(HSLish, Space):
    """HPLuv class."""

    BASE = 'lchuv'
    NAME = "hpluv"
    SERIALIZE = ("--hpluv",)
    CHANNELS = (
        Channel("h", 0.0, 360.0, bound=True, flags=FLG_ANGLE),
        Channel("p", 0.0, 100.0, bound=True),
        Channel("l", 0.0, 100.0, bound=True)
    )
    CHANNEL_ALIASES = {
        "hue": "h",
        "perpendiculars": "p",
        "lightness": "l"
    }
    WHITE = WHITES['2deg']['D65']

    def is_achromatic(self, coords: Vector) -> bool:
        """Check if color is achromatic."""

        return abs(coords[1]) < 1e-4 or coords[2] > (100 - 1e-7) or coords[2] < 1e-08

    def to_base(self, coords: Vector) -> Vector:
        """To LChuv from HPLuv."""

        return hpluv_to_lch(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From LChuv to HPLuv."""

        return lch_to_hpluv(coords)



"""
XYB color space.

https://ds.jpeg.org/whitepapers/jpeg-xl-whitepaper.pdf
"""
from __future__ import annotations
from .. import algebra as alg
from ..spaces import Space, Labish
from ..types import Vector
from ..cat import WHITES
from ..channels import Channel, FLG_MIRROR_PERCENT

BIAS = 0.00379307325527544933
BIAS_CBRT = alg.nth_root(BIAS, 3)

LRGB_TO_LMS = [
    [0.3, 0.622, 0.078],
    [0.23, 0.692, 0.078],
    [0.24342268924547819, 0.20476744424496821, 0.55180986650955360]
]

LMS_TO_LRGB = [
    [11.031566904639865, -9.866943908131564, -0.16462299650829934],
    [-3.254147381074425, 4.4187703775827245, -0.16462299650829929],
    [-3.6588512867136815, 2.712923045936092, 1.945928240777589]
]

# https://twitter.com/jonsneyers/status/1605321352143331328
# @jonsneyers Feb 22
# Yes, the default is to just subtract Y from B. In general there are locally
# signaled float multipliers to subtract some multiple of Y from X and some
# other multiple from B. But this is the baseline, making X=B=0 grayscale.
# ----
# We adjust the matrix to subtract Y from B match this statement.
XYB_LMS_TO_XYB = [
    [0.5, -0.5, 0.0],
    [0.5, 0.5, 0.0],
    [0.0, -1.0, 1.0],
]

XYB_TO_XYB_LMS = [
    [1.0, 1.0, 0.0],
    [-1.0, 1.0, 0.0],
    [-1.0, 1.0, 1.0]
]


def rgb_to_xyb(rgb: Vector) -> Vector:
    """Linear sRGB to XYB."""

    return alg.dot(
        XYB_LMS_TO_XYB,
        [alg.nth_root(c + BIAS, 3) - BIAS_CBRT for c in alg.dot(LRGB_TO_LMS, rgb, dims=alg.D2_D1)],
        dims=alg.D2_D1
    )


def xyb_to_rgb(xyb: Vector) -> Vector:
    """XYB to linear sRGB."""

    # This cleans up the round trip on black.
    if not any(xyb):
        return [0.0] * 3

    return alg.dot(
        LMS_TO_LRGB,
        [(c + BIAS_CBRT) ** 3 - BIAS for c in alg.dot(XYB_TO_XYB_LMS, xyb, dims=alg.D2_D1)],
        dims=alg.D2_D1
    )


class XYB(Labish, Space):
    """XYB color class."""

    BASE = 'srgb-linear'
    NAME = "xyb"
    SERIALIZE = ("--xyb",)
    WHITE = WHITES['2deg']['D65']
    CHANNELS = (
        Channel("x", -0.45, 0.45, flags=FLG_MIRROR_PERCENT),
        Channel("y", 0.0, 0.845),
        Channel("b", -0.45, 0.45, flags=FLG_MIRROR_PERCENT)
    )

    def names(self) -> tuple[str, ...]:
        """Return Lab-ish names in the order L a b."""

        channels = self.channels
        return channels[1], channels[0], channels[2]

    def to_base(self, coords: Vector) -> Vector:
        """To XYB from base."""

        return xyb_to_rgb(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From base to XYB."""

        return rgb_to_xyb(coords)



"""Model gamut mapping."""
import sys
import os
import matplotlib.pyplot as plt
import argparse

sys.path.insert(0, os.getcwd())

try:
    from coloraide_extras.everything import ColorAll as Color
except ImportError:
    from coloraide.everything import ColorAll as Color
from tools.slice_diagram import plot_slice  # noqa: E402
from coloraide.util import fmt_float  # noqa: E402


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='gamut_diagrams', description='Demonstrate gamut mapping.')
    parser.add_argument('--color', '-c', help="Color to gamut map.")
    parser.add_argument('--method', '-m', default='lch-chroma', help="Gamut map method")
    parser.add_argument('--gamut', '-g', default="srgb", help='Gamut to evaluate the color in (default is sRGB).')
    parser.add_argument('--no-border', '-b', action="store_true", help='Draw no border around the graphed content.')
    parser.add_argument('--resolution', '-r', default="500", help="How densely to render the figure.")
    parser.add_argument('--clip-space', '-p', default='lch', help="LCh space to show clipping in.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--dpi', default=200, type=int, help="DPI of image.")
    parser.add_argument('--output', '-o', default='', help='Output file.')

    args = parser.parse_args()
    method = args.method

    if method == 'clip':
        if args.clip_space not in ('lch', 'oklch', 'hct'):
            raise ValueError('"{}" is an unsupported clipping space'.format(args.clip_space))
        method = args.clip_space + '-chroma'

    if method == 'lch-chroma':
        space = 'lch'
        t_space = 'CIELCh'
        xaxis = 'c:0:264'
        yaxis = 'l:0:100'
        x = 'c'
        y = 'l'
    elif method == 'oklch-chroma':
        space = 'oklch'
        t_space = 'OkLCh'
        xaxis = 'c:0:1.5'
        yaxis = 'l:0:1'
        x = 'c'
        y = 'l'
    elif method == 'hct-chroma':
        space = 'hct'
        t_space = 'HCT'
        xaxis = 'c:0:267'
        yaxis = 't:0:100'
        x = 'c'
        y = 't'
    else:
        raise ValueError('"{}" is an unsupported gamut mapping algorithm'.format(args.method))

    if args.method == 'clip':
        title = 'Clipping shown in {}'.format(t_space)
    else:
        title = 'MINDE and Chroma Reduction in {}'.format(t_space)

    color = Color(args.color).convert(space, in_place=True)
    color2 = color.clone().fit('srgb', method=args.method)
    mapcolor = color.convert(space)
    mapcolor2 = color2.convert(space)
    constant = 'h:{}'.format(fmt_float(mapcolor['hue'], 5))
    subtitle = '{} ==> {}'.format(color.to_string(), color2.to_string())

    plot_slice(
        space,
        constant,
        xaxis,
        yaxis,
        gamut=args.gamut,
        resolution=int(args.resolution),
        title=title,
        subtitle=subtitle,
        dark=args.dark,
        border=not args.no_border
    )

    plt.plot(
        [mapcolor2[x], mapcolor[x]],
        [mapcolor2[y], mapcolor[y]],
        color='black',
        marker="",
        linewidth=1.5,
        markersize=2,
        antialiased=True
    )

    plt.scatter(
        mapcolor[x],
        mapcolor[y],
        marker="o",
        color=color2.convert('srgb').to_string(hex=True, fit=args.method),
        edgecolor='black',
        zorder=100
    )

    plt.scatter(
        mapcolor2[x],
        mapcolor2[y],
        marker="o",
        color=color2.convert('srgb').to_string(hex=True, fit=args.method),
        edgecolor='black',
        zorder=100
    )

    if args.output:
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.gcf().set_dpi(args.dpi)
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""Test Luv library."""
import unittest
from . import util
from coloraide.everything import ColorAll as Color, NaN
import pytest


class TestLuv(util.ColorAssertsPyTest):
    """Test Luv."""

    COLORS = [
        ('red', 'color(--luv 53.237 175.01 37.765)'),
        ('orange', 'color(--luv 74.934 74.839 74.014)'),
        ('yellow', 'color(--luv 97.139 7.7042 106.81)'),
        ('green', 'color(--luv 46.228 -43.768 56.599)'),
        ('blue', 'color(--luv 32.301 -9.4024 -130.35)'),
        ('indigo', 'color(--luv 20.47 10.08 -61.34)'),
        ('violet', 'color(--luv 69.695 51.84 -67.037)'),
        ('white', 'color(--luv 100 0 0)'),
        ('gray', 'color(--luv 53.585 0 0)'),
        ('black', 'color(--luv 0 0 0)'),
        # Test color
        ('color(--luv 100 0.1 -0.1)', 'color(--luv 100 0.1 -0.1)'),
        ('color(--luv 100 0.1 -0.1 / 0.5)', 'color(--luv 100 0.1 -0.1 / 0.5)'),
        ('color(--luv 50% 50% -50% / 50%)', 'color(--luv 50 107.5 -107.5 / 0.5)'),
        ('color(--luv none none none / none)', 'color(--luv none none none / none)'),
        # Test percent ranges
        ('color(--luv 0% 0% 0%)', 'color(--luv 0 0 0)'),
        ('color(--luv 100% 100% 100%)', 'color(--luv 100 215 215)'),
        ('color(--luv -100% -100% -100%)', 'color(--luv -100 -215 -215)')
    ]

    @pytest.mark.parametrize('color1,color2', COLORS)
    def test_colors(self, color1, color2):
        """Test colors."""

        self.assertColorEqual(Color(color1).convert('luv'), Color(color2))


class TestLuvSerialize(util.ColorAssertsPyTest):
    """Test Luv serialization."""

    COLORS = [
        # Test color
        ('color(--luv 75 10 -10 / 0.5)', {}, 'color(--luv 75 10 -10 / 0.5)'),
        # Test alpha
        ('color(--luv 75 10 -10)', {'alpha': True}, 'color(--luv 75 10 -10 / 1)'),
        ('color(--luv 75 10 -10 / 0.5)', {'alpha': False}, 'color(--luv 75 10 -10)'),
        # Test None
        ('color(--luv none 10 -10)', {}, 'color(--luv 0 10 -10)'),
        ('color(--luv none 10 -10)', {'none': True}, 'color(--luv none 10 -10)'),
        # Test Fit (not bound)
        ('color(--luv 120 10 -10)', {}, 'color(--luv 120 10 -10)'),
        ('color(--luv 120 10 -10)', {'fit': False}, 'color(--luv 120 10 -10)')
    ]

    @pytest.mark.parametrize('color1,options,color2', COLORS)
    def test_colors(self, color1, options, color2):
        """Test colors."""

        self.assertEqual(Color(color1).to_string(**options), color2)


class TestLuvProperties(util.ColorAsserts, unittest.TestCase):
    """Test Luv."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c['lightness'], 90)
        c['lightness'] = 80
        self.assertEqual(c['lightness'], 80)

    def test_u(self):
        """Test `u`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c['u'], 50)
        c['u'] = 40
        self.assertEqual(c['u'], 40)

    def test_v(self):
        """Test `v`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c['v'], -20)
        c['v'] = -10
        self.assertEqual(c['v'], -10)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--luv 90% 50 -20 / 1)')
        self.assertEqual(c['alpha'], 1)
        c['alpha'] = 0.5
        self.assertEqual(c['alpha'], 0.5)


class TestsAchromatic(util.ColorAsserts, unittest.TestCase):
    """Test achromatic."""

    def test_achromatic(self):
        """Test when color is achromatic."""

        self.assertEqual(Color('luv', [30, 0, 0]).is_achromatic(), True)
        self.assertEqual(Color('luv', [30, 0.000001, 0]).is_achromatic(), True)
        self.assertEqual(Color('luv', [NaN, 0.00001, 0]).is_achromatic(), True)
        self.assertEqual(Color('luv', [0, NaN, NaN]).is_achromatic(), True)
        self.assertEqual(Color('luv', [0, 30, -40]).is_achromatic(), True)
        self.assertEqual(Color('luv', [NaN, 0, -30]).is_achromatic(), True)
        self.assertEqual(Color('luv', [30, NaN, 0]).is_achromatic(), True)
        self.assertEqual(Color('luv', [NaN, NaN, 0]).is_achromatic(), True)



"""
ACEScc color space.

https://www.oscars.org/science-technology/aces/aces-documentation
"""
from __future__ import annotations
import math
from ..channels import Channel
from ..spaces.srgb import sRGB
from ..types import Vector

CC_MIN = (math.log2(2 ** -16) + 9.72) / 17.52
CC_MAX = (math.log2(65504) + 9.72) / 17.52


def acescc_to_acescg(acescc: Vector) -> Vector:
    """Convert ACEScc to XYZ."""

    c1 = 2 ** -16
    c2 = (9.72 - 15) / 17.52
    c3 = (math.log2(65504) + 9.72) / 17.52

    acescg = []
    for c in acescc:
        if c <= c2:
            c = (2 ** (c * 17.52 - 9.72) - c1) * 2.0
        elif c2 <= c < c3:
            c = 2 ** (c * 17.52 - 9.72)
        else:
            c = 65504.0
        acescg.append(c)
    return acescg


def acescg_to_acescc(acescg: Vector) -> Vector:
    """Convert XYZ to ACEScc."""

    c1 = 2 ** -16
    c2 = 2 ** -15

    acescc = []
    for c in acescg:
        if c <= 0:
            c = math.log2(c1)
        elif c < c2:
            c = math.log2(c1 + c * 0.5)
        else:
            c = math.log2(c)
        acescc.append((c + 9.72) / 17.52)
    return acescc


class ACEScc(sRGB):
    """The ACEScc color class."""

    BASE = "acescg"
    NAME = "acescc"
    SERIALIZE = ("--acescc",)  # type: tuple[str, ...]
    WHITE = (0.32168, 0.33767)
    CHANNELS = (
        Channel("r", CC_MIN, CC_MAX, bound=True, nans=CC_MIN),
        Channel("g", CC_MIN, CC_MAX, bound=True, nans=CC_MIN),
        Channel("b", CC_MIN, CC_MAX, bound=True, nans=CC_MIN)
    )
    DYNAMIC_RANGE = 'hdr'

    def to_base(self, coords: Vector) -> Vector:
        """To XYZ."""

        return acescc_to_acescg(coords)

    def from_base(self, coords: Vector) -> Vector:
        """From XYZ."""

        return acescg_to_acescc(coords)
